import logging

import numpy as np

from autofit import conf
from autofit.mapper import model_mapper as mm
from autofit.optimize.non_linear.paths import Paths, convert_paths

logging.basicConfig()
logger = logging.getLogger(__name__)  # TODO: Logging issue


class NonLinearOptimizer:
    @convert_paths
    def __init__(self, paths):
        """Abstract base class for non-linear optimizers.

        This class sets up the file structure for the non-linear optimizer nlo, which are standardized across \
        all non-linear optimizers.

        Parameters
        ------------

        """
        log_file = conf.instance.general.get("output", "log_file", str).replace(" ", "")
        self.paths = paths

        if not len(log_file) == 0:
            log_path = "{}/{}".format(self.paths.phase_output_path, log_file)
            logger.handlers = [logging.FileHandler(log_path)]
            logger.propagate = False
            # noinspection PyProtectedMember
            logger.level = logging._nameToLevel[
                conf.instance.general.get("output", "log_level", str)
                .replace(" ", "")
                .upper()
            ]

        self.paths.restore()

    def config(self, attribute_name, attribute_type=str):
        """
        Get a config field from this optimizer's section in non_linear.ini by a key and value type.

        Parameters
        ----------
        attribute_name: str
            The analysis_path of the field
        attribute_type: type
            The type of the value

        Returns
        -------
        attribute
            An attribute for the key with the specified type.
        """
        return conf.instance.non_linear.get(
            self.__class__.__name__, attribute_name, attribute_type
        )

    def __eq__(self, other):
        return isinstance(other, NonLinearOptimizer) and self.__dict__ == other.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.paths.restore()

    def fit(self, analysis, model):
        raise NotImplementedError(
            "Fitness function must be overridden by non linear optimizers"
        )

    class Fitness:
        def __init__(
            self, paths, analysis, output_results=lambda during_analysis: None
        ):
            self.output_results = output_results
            self.paths = paths
            self.result = None
            self.max_likelihood = -np.inf
            self.analysis = analysis

            self.log_interval = conf.instance.general.get("output", "log_interval", int)
            self.backup_interval = conf.instance.general.get(
                "output", "backup_interval", int
            )
            self.visualize_interval = conf.instance.visualize_general.get(
                "general", "visualize_interval", int
            )
            self.model_results_output_interval = conf.instance.general.get(
                "output", "model_results_output_interval", int
            )

            self.should_log = IntervalCounter(self.log_interval)
            self.should_backup = IntervalCounter(self.backup_interval)
            self.should_visualize = IntervalCounter(self.visualize_interval)
            self.should_output_model_results = IntervalCounter(
                self.model_results_output_interval
            )

        def fit_instance(self, instance):
            likelihood = self.analysis.fit(instance)

            if likelihood > self.max_likelihood:

                self.max_likelihood = likelihood
                self.result = Result(instance, likelihood)

                if self.should_visualize():
                    self.analysis.visualize(instance, during_analysis=True)

                if self.should_backup():
                    self.paths.backup()

                if self.should_output_model_results():
                    self.output_results(during_analysis=True)

            return likelihood

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        name = "{}/{}".format(self.paths.phase_name, extension)

        if remove_phase_tag:
            phase_tag = ""
        else:
            phase_tag = self.paths.phase_tag

        new_instance = self.__class__(
            Paths(
                phase_name=name,
                phase_folders=self.paths.phase_folders,
                phase_tag=phase_tag,
                remove_files=self.paths.remove_files,
            )
        )

        return new_instance

    def output_from_model(self, model, paths):
        raise NotImplementedError()


class Analysis:
    def fit(self, instance):
        raise NotImplementedError()

    def visualize(self, instance, during_analysis):
        raise NotImplementedError()


class Result:
    """
    @DynamicAttrs
    """

    def __init__(
        self, instance, figure_of_merit, previous_model=None, gaussian_tuples=None
    ):
        """
        The result of an optimization.

        Parameters
        ----------
        instance: autofit.mapper.model.ModelInstance
            An instance object comprising the class instances that gave the optimal fit
        figure_of_merit: float
            A value indicating the figure of merit given by the optimal fit
        previous_model: mm.ModelMapper
            The model mapper from the stage that produced this result
        """
        self.instance = instance
        self.figure_of_merit = figure_of_merit
        self.previous_model = previous_model
        self.gaussian_tuples = gaussian_tuples
        self.__model = None

    @property
    def model(self):
        if self.__model is None:
            self.__model = self.previous_model.mapper_from_gaussian_tuples(
                self.gaussian_tuples
            )
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(
                ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
            )
        )

    def model_absolute(self, a: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined absolute
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.gaussian_tuples, a=a
        )

    def model_relative(self, r: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined relative
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.gaussian_tuples, r=r
        )


class IntervalCounter:
    def __init__(self, interval):
        self.count = 0
        self.interval = interval

    def __call__(self):
        if self.interval == -1:
            return False
        self.count += 1
        return self.count % self.interval == 0



import os
import shutil
from functools import wraps

import pytest

import autofit as af
from autofit import Paths
from autofit.optimize.non_linear.output import AbstractOutput
from test_autofit.mock import MockClassNLOx4, MockClassNLOx5, MockClassNLOx6

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class MockOutput(AbstractOutput):
    def __init__(
        self,
        model,
        paths,
        most_probable_model_parameters=None,
        most_likely_model_parameters=None,
        model_parameters_at_sigma_limit=None,
        sample_model_parameters=None,
    ):

        super(MockOutput, self).__init__(model=model, paths=paths)

        self._most_probable_model_parameters = most_probable_model_parameters
        self._most_likely_model_parameters = most_likely_model_parameters
        self._model_parameters_at_sigma_limit = model_parameters_at_sigma_limit
        self._sample_model_parameters = sample_model_parameters

    @property
    def most_probable_model_parameters(self):
        return self._most_probable_model_parameters

    @property
    def most_likely_model_parameters(self):
        return self._most_likely_model_parameters

    def model_parameters_at_sigma_limit(self, sigma_limit):
        return [
            (sigma_limit * value[0], sigma_limit * value[1])
            for value in self._model_parameters_at_sigma_limit
        ]

    def sample_model_parameters_from_sample_index(self, sample_index):
        return self._sample_model_parameters[sample_index]


class TestOutput:
    def test__most_probable_model_instance(self):

        model = af.ModelMapper(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_model_parameters=[
                1.0,
                2.0,
                3.0,
                4.0,
                -5.0,
                -6.0,
                -7.0,
                -8.0,
                9.0,
                10.0,
            ],
        )

        most_probable = output.most_probable_model_instance

        assert most_probable.mock_class_1.one == 1.0
        assert most_probable.mock_class_1.two == 2.0
        assert most_probable.mock_class_1.three == 3.0
        assert most_probable.mock_class_1.four == 4.0

        assert most_probable.mock_class_2.one == (-5.0, -6.0)
        assert most_probable.mock_class_2.two == (-7.0, -8.0)
        assert most_probable.mock_class_2.three == 9.0
        assert most_probable.mock_class_2.four == 10.0

    def test__most_likely_model_instance(self):

        model = af.ModelMapper(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

        output = MockOutput(
            model=model,
            paths=Paths(),
            most_likely_model_parameters=[
                21.0,
                22.0,
                23.0,
                24.0,
                25.0,
                -26.0,
                -27.0,
                28.0,
                29.0,
                30.0,
            ],
        )

        most_likely = output.most_likely_model_instance

        assert most_likely.mock_class_1.one == 21.0
        assert most_likely.mock_class_1.two == 22.0
        assert most_likely.mock_class_1.three == 23.0
        assert most_likely.mock_class_1.four == 24.0

        assert most_likely.mock_class_2.one == (25.0, -26.0)
        assert most_likely.mock_class_2.two == (-27.0, 28.0)
        assert most_likely.mock_class_2.three == 29.0
        assert most_likely.mock_class_2.four == 30.0

    def test__model_parameters_at_upper_and_lower_sigma_limit(self,):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_model_parameters=[1.0, 2.0, 3.0, 4.1],
            model_parameters_at_sigma_limit=[
                (0.88, 1.12),
                (1.88, 2.12),
                (2.88, 3.12),
                (3.88, 4.12),
            ],
        )

        params_upper = output.model_parameters_at_upper_sigma_limit(sigma_limit=1.0)
        assert params_upper == pytest.approx([1.12, 2.12, 3.12, 4.12], 1e-2)
        params_lower = output.model_parameters_at_lower_sigma_limit(sigma_limit=1.0)
        assert params_lower == pytest.approx([0.88, 1.88, 2.88, 3.88], 1e-2)

        params_upper = output.model_parameters_at_upper_sigma_limit(sigma_limit=2.0)
        assert params_upper == pytest.approx(
            [2.0 * 1.12, 2.0 * 2.12, 2.0 * 3.12, 2.0 * 4.12], 1e-2
        )
        params_lower = output.model_parameters_at_lower_sigma_limit(sigma_limit=2.0)
        assert params_lower == pytest.approx(
            [2.0 * 0.88, 2.0 * 1.88, 2.0 * 2.88, 2.0 * 3.88], 1e-2
        )

    def test__gaussian_priors(self):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_model_parameters=[1.0, 2.0, 3.0, 4.1],
            model_parameters_at_sigma_limit=[
                (0.88, 1.12),
                (1.88, 2.12),
                (2.88, 3.12),
                (3.88, 4.12),
            ],
        )

        gaussian_priors = output.gaussian_priors_at_sigma_limit(sigma_limit=1.0)

        assert gaussian_priors[0][0] == 1.0
        assert gaussian_priors[1][0] == 2.0
        assert gaussian_priors[2][0] == 3.0
        assert gaussian_priors[3][0] == 4.1

        assert gaussian_priors[0][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[1][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[2][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[3][1] == pytest.approx(0.22, 1e-2)

    def test__values_offset_from_input_model_parameters(self):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_model_parameters=[1.0, -2.0, 3.0, 4.0],
        )

        offset_values = output.values_offset_from_input_model_parameters(
            input_model_parameters=[1.0, 1.0, 2.0, 3.0]
        )

        assert offset_values == [0.0, -3.0, 1.0, 1.0]

    def test__sample_model_instance_from_sample_index(self,):

        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            sample_model_parameters=[[1.0, -2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]],
        )

        instance = output.sample_model_instance_from_sample_index(sample_index=0)

        assert instance.mock_class.one == 1.0
        assert instance.mock_class.two == -2.0
        assert instance.mock_class.three == 3.0
        assert instance.mock_class.four == 4.0

        instance = output.sample_model_instance_from_sample_index(sample_index=1)

        assert instance.mock_class.one == 5.0
        assert instance.mock_class.two == 6.0
        assert instance.mock_class.three == 7.0
        assert instance.mock_class.four == 8.0

    def test__model_errors_at_sigma_limit(self,):
        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            model_parameters_at_sigma_limit=[
                (0.88, 1.12),
                (1.88, 2.12),
                (2.88, 3.12),
                (3.88, 4.12),
            ],
        )

        model_errors = output.model_errors_at_sigma_limit(sigma_limit=1.0)
        assert model_errors == pytest.approx(
            [1.12 - 0.88, 2.12 - 1.88, 3.12 - 2.88, 4.12 - 3.88], 1e-2
        )

        model_errors_instance = output.model_errors_instance_at_sigma_limit(
            sigma_limit=1.0
        )
        assert model_errors_instance.mock_class.one == pytest.approx(1.12 - 0.88, 1e-2)
        assert model_errors_instance.mock_class.two == pytest.approx(2.12 - 1.88, 1e-2)
        assert model_errors_instance.mock_class.three == pytest.approx(
            3.12 - 2.88, 1e-2
        )
        assert model_errors_instance.mock_class.four == pytest.approx(4.12 - 3.88, 1e-2)

        model_errors = output.model_errors_at_sigma_limit(sigma_limit=2.0)
        assert model_errors == pytest.approx(
            [
                2.0 * (1.12 - 0.88),
                2.0 * (2.12 - 1.88),
                2.0 * (3.12 - 2.88),
                2.0 * (4.12 - 3.88),
            ],
            1e-2,
        )

    def test__model_errors_at_upper_and_lower_sigma_limit(self,):
        model = af.ModelMapper(mock_class=MockClassNLOx4)
        output = MockOutput(
            model=model,
            paths=Paths(),
            most_probable_model_parameters=[1.1, 2.0, 3.0, 4.0],
            model_parameters_at_sigma_limit=[
                (0.88, 1.12),
                (1.88, 2.12),
                (2.88, 3.12),
                (3.88, 4.12),
            ],
        )

        model_upper_errors = output.model_errors_at_upper_sigma_limit(sigma_limit=1.0)
        assert model_upper_errors == pytest.approx([0.02, 0.12, 0.12, 0.12], 1e-2)

        model_errors_instance = output.model_errors_instance_at_upper_sigma_limit(
            sigma_limit=1.0
        )
        assert model_errors_instance.mock_class.one == pytest.approx(0.02, 1e-2)
        assert model_errors_instance.mock_class.two == pytest.approx(0.12, 1e-2)
        assert model_errors_instance.mock_class.three == pytest.approx(0.12, 1e-2)
        assert model_errors_instance.mock_class.four == pytest.approx(0.12, 1e-2)

        model_upper_errors = output.model_errors_at_upper_sigma_limit(sigma_limit=2.0)
        assert model_upper_errors == pytest.approx([1.14, 2.24, 3.24, 4.24], 1e-2)

        model_lower_errors = output.model_errors_at_lower_sigma_limit(sigma_limit=1.0)
        assert model_lower_errors == pytest.approx([0.22, 0.12, 0.12, 0.12], 1e-2)

        model_errors_instance = output.model_errors_instance_at_lower_sigma_limit(
            sigma_limit=1.0
        )
        assert model_errors_instance.mock_class.one == pytest.approx(0.22, 1e-2)
        assert model_errors_instance.mock_class.two == pytest.approx(0.12, 1e-2)
        assert model_errors_instance.mock_class.three == pytest.approx(0.12, 1e-2)
        assert model_errors_instance.mock_class.four == pytest.approx(0.12, 1e-2)

        model_lower_errors = output.model_errors_at_lower_sigma_limit(sigma_limit=0.5)
        assert model_lower_errors == pytest.approx([0.66, 1.06, 1.56, 2.06], 1e-2)



"""
This package implements the No-U-Turn Sampler (NUTS) algorithm 6 from the NUTS
paper (Hoffman & Gelman, 2011).

Content
-------

The package mainly contains:
  nuts6                     return samples using the NUTS
  test_nuts6                example usage of this package

and subroutines of nuts6:
  build_tree                the main recursion in NUTS
  find_reasonable_epsilon   Heuristic for choosing an initial value of epsilon
  leapfrog                  Perfom a leapfrog jump in the Hamiltonian space
  stop_criterion            Compute the stop condition in the main loop


A few words about NUTS
----------------------

Hamiltonian Monte Carlo or Hybrid Monte Carlo (HMC) is a Markov chain Monte
Carlo (MCMC) algorithm that avoids the random walk behavior and sensitivity to
correlated parameters, biggest weakness of many MCMC methods. Instead, it takes
a series of steps informed by first-order gradient information.

This feature allows it to converge much more quickly to high-dimensional target
distributions compared to simpler methods such as Metropolis, Gibbs sampling
(and derivatives).

However, HMC's performance is highly sensitive to two user-specified
parameters: a step size, and a desired number of steps.  In particular, if the
number of steps is too small then the algorithm will just exhibit random walk
behavior, whereas if it is too large it will waste computations.

Hoffman & Gelman introduced NUTS or the No-U-Turn Sampler, an extension to HMC
that eliminates the need to set a number of steps.  NUTS uses a recursive
algorithm to find likely candidate points that automatically stops when it
starts to double back and retrace its steps.  Empirically, NUTS perform at
least as effciently as and sometimes more effciently than a well tuned standard
HMC method, without requiring user intervention or costly tuning runs.

Moreover, Hoffman & Gelman derived a method for adapting the step size
parameter on the fly based on primal-dual averaging.  NUTS can thus be used
with no hand-tuning at all.

In practice, the implementation still requires a number of steps, a burning
period and a stepsize. However, the stepsize will be optimized during the
burning period, and the final values of all the user-defined values will be
revised by the algorithm.

reference: arXiv:1111.4246
"The No-U-Turn Sampler: Adaptively Setting Path Lengths in Hamiltonian Monte
Carlo", Matthew D. Hoffman & Andrew Gelman
"""
import numpy as np
from numpy import log, exp, sqrt

__all__ = ["nuts6"]


def leapfrog(theta, r, grad, epsilon, f):
    """ Perfom a leapfrog jump in the Hamiltonian space
    INPUTS
    ------
    theta: ndarray[float, ndim=1]
        initial parameter position

    r: ndarray[float, ndim=1]
        initial momentum

    grad: float
        initial gradient value

    epsilon: float
        step size

    f: callable
        it should return the log probability and gradient evaluated at theta
        logp, grad = f(theta)

    OUTPUTS
    -------
    thetaprime: ndarray[float, ndim=1]
        new parameter position
    rprime: ndarray[float, ndim=1]
        new momentum
    gradprime: float
        new gradient
    logpprime: float
        new lnp
    """
    # make half step in r
    rprime = r + 0.5 * epsilon * grad
    # make new step in theta
    thetaprime = theta + epsilon * rprime
    # compute new gradient
    logpprime, gradprime = f(thetaprime)
    # make half step in r again
    rprime = rprime + 0.5 * epsilon * gradprime
    return thetaprime, rprime, gradprime, logpprime


def find_reasonable_epsilon(theta0, grad0, logp0, f):
    """ Heuristic for choosing an initial value of epsilon """
    epsilon = 1.0
    r0 = np.random.normal(0.0, 1.0, len(theta0))

    # Figure out what direction we should be moving epsilon.
    _, rprime, gradprime, logpprime = leapfrog(theta0, r0, grad0, epsilon, f)
    # brutal! This trick make sure the step is not huge leading to infinite
    # values of the likelihood. This could also help to make sure theta stays
    # within the prior domain (if any)
    k = 1.0
    while np.isinf(logpprime) or np.isinf(gradprime).any():
        k *= 0.5
        _, rprime, _, logpprime = leapfrog(theta0, r0, grad0, epsilon * k, f)

    epsilon = 0.5 * k * epsilon

    # acceptprob = np.exp(logpprime - logp0 - 0.5 * (np.dot(rprime, rprime.T) - np.dot(r0, r0.T)))
    # a = 2. * float((acceptprob > 0.5)) - 1.
    logacceptprob = logpprime - logp0 - 0.5 * (np.dot(rprime, rprime) - np.dot(r0, r0))
    a = 1.0 if logacceptprob > np.log(0.5) else -1.0
    # Keep moving epsilon in that direction until acceptprob crosses 0.5.
    # while ( (acceptprob ** a) > (2. ** (-a))):
    while a * logacceptprob > -a * np.log(2):
        epsilon = epsilon * (2.0 ** a)
        _, rprime, _, logpprime = leapfrog(theta0, r0, grad0, epsilon, f)
        # acceptprob = np.exp(logpprime - logp0 - 0.5 * ( np.dot(rprime, rprime.T) - np.dot(r0, r0.T)))
        logacceptprob = (
            logpprime - logp0 - 0.5 * (np.dot(rprime, rprime) - np.dot(r0, r0))
        )

    print("find_reasonable_epsilon=", epsilon)

    return epsilon


def stop_criterion(thetaminus, thetaplus, rminus, rplus):
    """ Compute the stop condition in the main loop
    dot(dtheta, rminus) >= 0 & dot(dtheta, rplus >= 0)

    INPUTS
    ------
    thetaminus, thetaplus: ndarray[float, ndim=1]
        under and above position
    rminus, rplus: ndarray[float, ndim=1]
        under and above momentum

    OUTPUTS
    -------
    criterion: bool
        return if the condition is valid
    """
    dtheta = thetaplus - thetaminus
    return (np.dot(dtheta, rminus.T) >= 0) & (np.dot(dtheta, rplus.T) >= 0)


def build_tree(theta, r, grad, logu, v, j, epsilon, f, joint0):
    """The main recursion."""
    if j == 0:
        # Base case: Take a single leapfrog step in the direction v.
        thetaprime, rprime, gradprime, logpprime = leapfrog(
            theta, r, grad, v * epsilon, f
        )
        joint = logpprime - 0.5 * np.dot(rprime, rprime.T)
        # Is the new point in the slice?
        nprime = int(logu < joint)
        # Is the simulation wildly inaccurate?
        sprime = int((logu - 1000.0) < joint)
        # Set the return values---minus=plus for all things here, since the
        # "tree" is of depth 0.
        thetaminus = thetaprime[:]
        thetaplus = thetaprime[:]
        rminus = rprime[:]
        rplus = rprime[:]
        gradminus = gradprime[:]
        gradplus = gradprime[:]
        # Compute the acceptance probability.
        alphaprime = min(1.0, np.exp(joint - joint0))
        # alphaprime = min(1., np.exp(logpprime - 0.5 * np.dot(rprime, rprime.T) - joint0))
        nalphaprime = 1
    else:
        # Recursion: Implicitly build the height j-1 left and right subtrees.
        thetaminus, rminus, gradminus, thetaplus, rplus, gradplus, thetaprime, gradprime, logpprime, nprime, sprime, alphaprime, nalphaprime = build_tree(
            theta, r, grad, logu, v, j - 1, epsilon, f, joint0
        )
        # No need to keep going if the stopping criteria were met in the first subtree.
        if sprime == 1:
            if v == -1:
                thetaminus, rminus, gradminus, _, _, _, thetaprime2, gradprime2, logpprime2, nprime2, sprime2, alphaprime2, nalphaprime2 = build_tree(
                    thetaminus, rminus, gradminus, logu, v, j - 1, epsilon, f, joint0
                )
            else:
                _, _, _, thetaplus, rplus, gradplus, thetaprime2, gradprime2, logpprime2, nprime2, sprime2, alphaprime2, nalphaprime2 = build_tree(
                    thetaplus, rplus, gradplus, logu, v, j - 1, epsilon, f, joint0
                )
            # Choose which subtree to propagate a sample up from.
            if np.random.uniform() < (
                float(nprime2) / max(float(int(nprime) + int(nprime2)), 1.0)
            ):
                thetaprime = thetaprime2[:]
                gradprime = gradprime2[:]
                logpprime = logpprime2
            # Update the number of valid points.
            nprime = int(nprime) + int(nprime2)
            # Update the stopping criterion.
            sprime = int(
                sprime
                and sprime2
                and stop_criterion(thetaminus, thetaplus, rminus, rplus)
            )
            # Update the acceptance probability statistics.
            alphaprime = alphaprime + alphaprime2
            nalphaprime = nalphaprime + nalphaprime2

    return (
        thetaminus,
        rminus,
        gradminus,
        thetaplus,
        rplus,
        gradplus,
        thetaprime,
        gradprime,
        logpprime,
        nprime,
        sprime,
        alphaprime,
        nalphaprime,
    )


def nuts6(f, M, Madapt, theta0, delta=0.6):
    """
    Implements the No-U-Turn Sampler (NUTS) algorithm 6 from from the NUTS
    paper (Hoffman & Gelman, 2011).

    Runs Madapt steps of burn-in, during which it adapts the step size
    parameter epsilon, then starts generating samples to return.

    Note the initial step size is tricky and not exactly the one from the
    initial paper.  In fact the initial step size could be given by the user in
    order to avoid potential problems

    INPUTS
    ------
    epsilon: float
        step size
        see nuts8 if you want to avoid tuning this parameter

    f: callable
        it should return the log probability and gradient evaluated at theta
        logp, grad = f(theta)

    M: int
        number of samples to generate.

    Madapt: int
        the number of steps of burn-in/how long to run the dual averaging
        algorithm to fit the step size epsilon.

    theta0: ndarray[float, ndim=1]
        initial guess of the parameters.

    KEYWORDS
    --------
    delta: float
        targeted acceptance fraction

    OUTPUTS
    -------
    samples: ndarray[float, ndim=2]
    M x D matrix of samples generated by NUTS.
    note: samples[0, :] = theta0
    """

    if len(np.shape(theta0)) > 1:
        raise ValueError("theta0 is expected to be a 1-D array")

    D = len(theta0)
    samples = np.empty((M + Madapt, D), dtype=float)
    lnprob = np.empty(M + Madapt, dtype=float)

    logp, grad = f(theta0)
    samples[0, :] = theta0
    lnprob[0] = logp

    # Choose a reasonable first epsilon by a simple heuristic.
    epsilon = find_reasonable_epsilon(theta0, grad, logp, f)

    # Parameters to the dual averaging algorithm.
    gamma = 0.05
    t0 = 10
    kappa = 0.75
    mu = log(10.0 * epsilon)

    # Initialize dual averaging algorithm.
    epsilonbar = 1
    Hbar = 0

    for m in range(1, M + Madapt):
        # Resample momenta.
        r0 = np.random.normal(0, 1, D)

        # joint lnp of theta and momentum r
        joint = logp - 0.5 * np.dot(r0, r0.T)

        # Resample u ~ uniform([0, exp(joint)]).
        # Equivalent to (log(u) - joint) ~ exponential(1).
        logu = float(joint - np.random.exponential(1, size=1))

        # if all fails, the next sample will be the previous one
        samples[m, :] = samples[m - 1, :]
        lnprob[m] = lnprob[m - 1]

        # initialize the tree
        thetaminus = samples[m - 1, :]
        thetaplus = samples[m - 1, :]
        rminus = r0[:]
        rplus = r0[:]
        gradminus = grad[:]
        gradplus = grad[:]

        j = 0  # initial heigth j = 0
        n = 1  # Initially the only valid point is the initial point.
        s = 1  # Main loop: will keep going until s == 0.

        while s == 1:
            # Choose a direction. -1 = backwards, 1 = forwards.
            v = int(2 * (np.random.uniform() < 0.5) - 1)

            # Double the size of the tree.
            if v == -1:
                thetaminus, rminus, gradminus, _, _, _, thetaprime, gradprime, logpprime, nprime, sprime, alpha, nalpha = build_tree(
                    thetaminus, rminus, gradminus, logu, v, j, epsilon, f, joint
                )
            else:
                _, _, _, thetaplus, rplus, gradplus, thetaprime, gradprime, logpprime, nprime, sprime, alpha, nalpha = build_tree(
                    thetaplus, rplus, gradplus, logu, v, j, epsilon, f, joint
                )

            # Use Metropolis-Hastings to decide whether or not to move to a
            # point from the half-tree we just generated.
            _tmp = min(1, float(nprime) / float(n))
            if (sprime == 1) and (np.random.uniform() < _tmp):
                samples[m, :] = thetaprime[:]
                lnprob[m] = logpprime
                logp = logpprime
                grad = gradprime[:]
            # Update number of valid points we've seen.
            n += nprime
            # Decide if it's time to stop.
            s = sprime and stop_criterion(thetaminus, thetaplus, rminus, rplus)
            # Increment depth.
            j += 1

        # Do adaptation of epsilon if we're still doing burn-in.
        eta = 1.0 / float(m + t0)
        Hbar = (1.0 - eta) * Hbar + eta * (delta - alpha / float(nalpha))
        if m <= Madapt:
            epsilon = exp(mu - sqrt(m) / gamma * Hbar)
            eta = m ** -kappa
            epsilonbar = exp((1.0 - eta) * log(epsilonbar) + eta * log(epsilon))
        else:
            epsilon = epsilonbar
    samples = samples[Madapt:, :]
    lnprob = lnprob[Madapt:]
    return samples, lnprob, epsilon



import os
import pytest

import autofit as af
from autofit import Paths
from autofit.optimize.non_linear.emcee import EmceeOutput
from test_autofit.mock import MockClassNLOx4

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="emcee_output")
def test_emcee_output():
    emcee_output_path = "{}/../test_files/non_linear/emcee/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    af.conf.instance.output_path = emcee_output_path

    mapper = af.ModelMapper(mock_class_1=MockClassNLOx4)

    return EmceeOutput(
        mapper,
        Paths(),
        auto_correlation_check_size=10,
        auto_correlation_required_length=10,
        auto_correlation_change_threshold=0.01,
    )


class TestEmceeOutput:
    def test__maximum_log_likelihood(self, emcee_output):

        assert emcee_output.maximum_log_likelihood == pytest.approx(583.26625, 1.0e-4)

    def test__most_probable_parameters(sel, emcee_output):

        assert emcee_output.most_probable_model_parameters == pytest.approx(
            [0.008422, -0.026413, 9.9579656, 0.494618], 1.0e-3
        )

    def test__most_likely_parameters(self, emcee_output):

        assert emcee_output.most_likely_model_parameters == pytest.approx(
            [0.003825, -0.00360509, 9.957799, 0.4940334], 1.0e-3
        )

    def test__model_parameters_at_sigma_limit__uses_output_files(self, emcee_output):

        params = emcee_output.model_parameters_at_sigma_limit(sigma_limit=3.0)

        assert params[0][0:2] == pytest.approx((-0.003197, 0.019923), 1e-2)

        params = emcee_output.model_parameters_at_sigma_limit(sigma_limit=1.0)

        assert params[0][0:2] == pytest.approx((0.0042278, 0.01087681), 1e-2)

    def test__samples__total_steps_samples__model_parameters_weight_and_likelihood_from_sample_index(
        self, emcee_output
    ):

        model = emcee_output.sample_model_parameters_from_sample_index(sample_index=0)
        weight = emcee_output.sample_weight_from_sample_index(sample_index=0)
        likelihood = emcee_output.sample_likelihood_from_sample_index(sample_index=0)

        assert emcee_output.total_walkers == 10
        assert emcee_output.total_steps == 1000
        assert emcee_output.total_samples == 10000
        assert model == pytest.approx(
            [0.0090338, -0.05790179, 10.192579, 0.480606], 1.0e-2
        )
        assert weight == 1.0
        assert likelihood == pytest.approx(-17257775239, 1.0e-4)

        #
        # assert emcee_output.total_samples == 10
        # assert model == [1.0, 2.0, 3.0, 4.0]
        # assert weight == 0.1
        # assert likelihood == -0.5 * 9999999.9

    def test__autocorrelation_times(self, emcee_output):

        assert (
            emcee_output.previous_auto_correlation_times_of_parameters
            == pytest.approx([31.92692, 36.54546, 73.33737, 67.52170], 1.0e-4)
        )
        assert emcee_output.auto_correlation_times_of_parameters == pytest.approx(
            [31.98507, 36.51001, 73.47629, 67.67495], 1.0e-4
        )

        assert emcee_output.converged == True



import pytest

import autofit as af
from autofit.mapper.prior_model import assertion as a


@pytest.fixture(name="prior_1")
def make_prior_1():
    return af.UniformPrior()


@pytest.fixture(name="prior_2")
def make_prior_2():
    return af.UniformPrior()


@pytest.fixture(name="lower_assertion")
def make_lower_assertion(prior_1, prior_2):
    return prior_1 < prior_2


@pytest.fixture(name="greater_assertion")
def make_greater_assertion(prior_1, prior_2):
    return prior_1 > prior_2


class TestAssertion:
    def test_lower_equal_assertion(self, prior_1, prior_2):
        assertion = prior_1 <= prior_2
        assertion({prior_1: 0.4, prior_2: 0.5})
        assertion({prior_1: 0.5, prior_2: 0.5})
        with pytest.raises(af.exc.FitException):
            assertion({prior_1: 0.6, prior_2: 0.5})

    def test_greater_equal_assertion(self, prior_1, prior_2):
        assertion = prior_1 >= prior_2
        assertion({prior_1: 0.6, prior_2: 0.5})
        assertion({prior_1: 0.5, prior_2: 0.5})
        with pytest.raises(af.exc.FitException):
            assertion({prior_1: 0.4, prior_2: 0.5})

    def test_lower_assertion(self, lower_assertion, prior_1, prior_2):
        assert isinstance(lower_assertion, a.GreaterThanLessThanAssertion)

        assert lower_assertion._lower is prior_1
        assert lower_assertion._greater is prior_2

    def test_greater_assertion(self, greater_assertion, prior_1, prior_2):
        assert isinstance(greater_assertion, a.GreaterThanLessThanAssertion)

        assert greater_assertion._lower is prior_2
        assert greater_assertion._greater is prior_1

    def test_assert_on_arguments_lower(self, lower_assertion, prior_1, prior_2):
        lower_assertion({prior_1: 0.3, prior_2: 0.5})
        with pytest.raises(af.exc.FitException):
            lower_assertion({prior_1: 0.6, prior_2: 0.5})

    def test_assert_on_arguments_greater(self, greater_assertion, prior_1, prior_2):
        greater_assertion({prior_1: 0.6, prior_2: 0.5})
        with pytest.raises(af.exc.FitException):
            greater_assertion({prior_1: 0.3, prior_2: 0.5})

    def test_numerical_assertion(self, prior_1):
        assertion = prior_1 < 0.5

        assertion({prior_1: 0.4})
        with pytest.raises(af.exc.FitException):
            assertion({prior_1: 0.6})

    def test_numerical_assertion_left(self, prior_1):
        assertion = 0.5 < prior_1

        assertion({prior_1: 0.6})
        with pytest.raises(af.exc.FitException):
            assertion({prior_1: 0.4})

    def test_compound_assertion(self, prior_1):
        assertion = (0.2 < prior_1) < 0.5
        assertion({prior_1: 0.3})
        with pytest.raises(af.exc.FitException):
            assertion({prior_1: 0.1})
        with pytest.raises(af.exc.FitException):
            assertion({prior_1: 0.6})


def test_assertion_in_model(prior_1, prior_2):
    model = af.ModelMapper()
    model.one = prior_1
    model.two = prior_2

    model.add_assertion(prior_1 < prior_2)

    model.instance_from_unit_vector([0.1, 0.2])
    with pytest.raises(af.exc.FitException):
        model.instance_from_unit_vector([0.2, 0.1])



import inspect
import math
import sys
from abc import ABC, abstractmethod
from typing import Union, Tuple

import numpy as np
from scipy.special import erfcinv

from autoconf import conf
from autofit import exc
from autofit.mapper.model_object import ModelObject
from autofit.mapper.prior_model.assertion import (
    GreaterThanLessThanAssertion,
    GreaterThanLessThanEqualAssertion,
)
from autofit.mapper.prior_model.attribute_pair import (
    cast_collection,
    PriorNameValue,
    InstanceNameValue,
)
from autofit.mapper.prior_model.deferred import DeferredArgument


class WidthModifier:
    def __init__(self, value):
        self.value = float(value)

    @classmethod
    def name_of_class(cls) -> str:
        """
        A string name for the class, with the prior suffix removed.
        """
        return cls.__name__.replace("WidthModifier", "")

    @classmethod
    def from_dict(cls, width_modifier_dict):
        return width_modifier_type_dict[width_modifier_dict["type"]](
            value=width_modifier_dict["value"]
        )

    @property
    def dict(self):
        return {"type": self.name_of_class(), "value": self.value}

    @staticmethod
    def for_class_and_attribute_name(cls, attribute_name):
        prior_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name, "width_modifier"]
        )
        return WidthModifier.from_dict(prior_dict)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.value == other.value


class RelativeWidthModifier(WidthModifier):
    def __call__(self, mean):
        return self.value * mean


class AbsoluteWidthModifier(WidthModifier):
    def __call__(self, _):
        return self.value


class TuplePrior:
    """
    A prior comprising one or more priors in a tuple
    """

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Prior)]
            A list of priors contained in this tuple
        """
        return list(filter(lambda t: isinstance(t[1], Prior), self.__dict__.items()))

    @property
    def unique_prior_tuples(self):
        return self.prior_tuples

    @property
    @cast_collection(InstanceNameValue)
    def instance_tuples(self):
        """
        Returns
        -------
        instances: [(String, instance)]
            A list of instances
        """
        return list(
            sorted(
                filter(lambda t: isinstance(t[1], float), self.__dict__.items()),
                key=lambda tup: tup[0],
            )
        )

    def value_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        tuple: (float,...)
            A tuple of float values
        """

        def convert(tup):
            if hasattr(tup, "prior"):
                return arguments[tup.prior]
            return tup.instance

        return tuple(
            map(
                convert,
                sorted(
                    self.prior_tuples + self.instance_tuples, key=lambda tup: tup.name
                ),
            )
        )

    def gaussian_tuple_prior_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        tuple_prior: TuplePrior
            A new tuple prior with gaussian priors
        """
        tuple_prior = TuplePrior()
        for prior_tuple in self.prior_tuples:
            setattr(tuple_prior, prior_tuple.name, arguments[prior_tuple.prior])
        return tuple_prior

    def __getitem__(self, item):
        return self.prior_tuples[item][1]


class Prior(ModelObject, ABC):
    def __init__(self, lower_limit=0.0, upper_limit=1.0):
        """
        An object used to mappers a unit value to an attribute value for a specific
        class attribute.

        Parameters
        ----------
        lower_limit: Float
            The lowest value this prior can return
        upper_limit: Float
            The highest value this prior can return
        """
        super().__init__()
        self.lower_limit = float(lower_limit)
        self.upper_limit = float(upper_limit)
        if self.lower_limit >= self.upper_limit:
            raise exc.PriorException(
                "The upper limit of a prior must be greater than its lower limit"
            )

    def assert_within_limits(self, value):
        if not (self.lower_limit <= value <= self.upper_limit):
            raise exc.PriorLimitException(
                "The physical value {} for a prior "
                "was not within its limits {}, {}".format(
                    value, self.lower_limit, self.upper_limit
                )
            )

    @staticmethod
    def for_class_and_attribute_name(cls, attribute_name):
        prior_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name]
        )
        return Prior.from_dict(prior_dict)

    @property
    def width(self):
        return self.upper_limit - self.lower_limit

    @abstractmethod
    def value_for(self, unit: float) -> float:
        """
        Return a physical value for a value between 0 and 1 with the transformation
        described by this prior.

        Parameters
        ----------
        unit
            A hypercube value between 0 and 1.

        Returns
        -------
        A physical value.
        """

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    def __gt__(self, other_prior: "Prior") -> GreaterThanLessThanAssertion:
        """
        Add an assertion that values associated with this prior are greater.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            lower physical values.

        Returns
        -------
        An assertion object
        """
        return GreaterThanLessThanAssertion(greater=self, lower=other_prior)

    def __lt__(self, other_prior: "Prior") -> GreaterThanLessThanAssertion:
        """
        Add an assertion that values associated with this prior are lower.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            greater physical values.

        Returns
        -------
        An assertion object
        """
        return GreaterThanLessThanAssertion(lower=self, greater=other_prior)

    def __ge__(self, other_prior: "Prior") -> GreaterThanLessThanEqualAssertion:
        """
        Add an assertion that values associated with this prior are greater or equal.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            lower physical values.

        Returns
        -------
        An assertion object
        """
        return GreaterThanLessThanEqualAssertion(greater=self, lower=other_prior)

    def __le__(self, other_prior: "Prior") -> GreaterThanLessThanEqualAssertion:
        """
        Add an assertion that values associated with this prior are lower or equal.

        Parameters
        ----------
        other_prior
            Another prior which is associated with a field that should always have
            greater physical values.

        Returns
        -------
        An assertion object
        """
        return GreaterThanLessThanEqualAssertion(lower=self, greater=other_prior)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<{} id={} lower_limit={} upper_limit={}>".format(
            self.__class__.__name__, self.id, self.lower_limit, self.upper_limit
        )

    @classmethod
    def from_dict(cls, prior_dict: dict) -> Union["Prior", DeferredArgument]:
        """
        Create a prior from a JSON representation

        Parameters
        ----------
        prior_dict
            A dictionary representation of a prior including a type (e.g. Uniform)
            and all constructor arguments.

        Returns
        -------
        An instance of a child of this class.
        """
        if prior_dict["type"] == "Constant":
            return prior_dict["value"]
        if prior_dict["type"] == "Deferred":
            return DeferredArgument()

        # noinspection PyProtectedMember
        return prior_type_dict[prior_dict["type"]](
            **{
                key: value
                for key, value in prior_dict.items()
                if key not in ("type", "width_modifier")
            }
        )

    @property
    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = {
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit,
            "type": self.name_of_class(),
        }
        return prior_dict

    @classmethod
    def name_of_class(cls) -> str:
        """
        A string name for the class, with the prior suffix removed.
        """
        return cls.__name__.replace("Prior", "")

    @property
    def limits(self) -> Tuple[float, float]:
        return self.lower_limit, self.upper_limit


class GaussianPrior(Prior):
    """A prior with a gaussian distribution"""

    def __init__(self, mean, sigma, lower_limit=-math.inf, upper_limit=math.inf):
        super().__init__(lower_limit, upper_limit)
        self.mean = float(mean)
        self.sigma = float(sigma)

    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute biased to the gaussian distribution
        """
        return self.mean + (self.sigma * math.sqrt(2) * erfcinv(2.0 * (1.0 - unit)))

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return (
            "GaussianPrior, mean = " + str(self.mean) + ", sigma = " + str(self.sigma)
        )

    def __repr__(self):
        return (
            "<GaussianPrior id={} mean={} sigma={} "
            "lower_limit={} upper_limit={}>".format(
                self.id, self.mean, self.sigma, self.lower_limit, self.upper_limit
            )
        )

    @property
    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = super().dict
        return {**prior_dict, "mean": self.mean, "sigma": self.sigma}


class UniformPrior(Prior):
    """A prior with a uniform distribution between a lower and upper limit"""

    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute between the upper and lower limits
        """
        return self.lower_limit + unit * (self.upper_limit - self.lower_limit)

    @property
    def mean(self):
        return self.lower_limit + (self.upper_limit - self.lower_limit) / 2

    @mean.setter
    def mean(self, new_value):
        difference = new_value - self.mean
        self.lower_limit += difference
        self.upper_limit += difference

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return (
            "UniformPrior, lower_limit = "
            + str(self.lower_limit)
            + ", upper_limit = "
            + str(self.upper_limit)
        )


class LogUniformPrior(UniformPrior):
    """A prior with a uniform distribution between a lower and upper limit"""

    def value_for(self, unit):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1
        Returns
        -------
        value: Float
            A value for the attribute between the upper and lower limits
        """
        return 10.0 ** (
            np.log10(self.lower_limit)
            + unit * (np.log10(self.upper_limit) - np.log10(self.lower_limit))
        )

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return (
            "LogUniformPrior, lower_limit = "
            + str(self.lower_limit)
            + ", upper_limit = "
            + str(self.upper_limit)
        )


def make_type_dict(cls):
    return {
        obj.name_of_class(): obj
        for _, obj in inspect.getmembers(sys.modules[__name__])
        if (inspect.isclass(obj) and issubclass(obj, cls) and obj != Prior)
    }


prior_type_dict = make_type_dict(Prior)

width_modifier_type_dict = make_type_dict(WidthModifier)



import glob
import logging
import os
import shutil
import zipfile
from functools import wraps

from autofit import conf
from autofit.mapper import link

logger = logging.getLogger(__name__)


def make_path(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        full_path = func(*args, **kwargs)
        if not os.path.exists(full_path):
            try:
                os.makedirs(full_path)
            except FileExistsError:
                pass
        return full_path

    return wrapper


def convert_paths(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(args) > 1:
            raise AssertionError(
                "Only phase name is allowed to be a positional argument in a phase constructor"
            )

        first_arg = kwargs.pop("paths", None)
        if first_arg is None and len(args) == 1:
            first_arg = args[0]

        if isinstance(first_arg, Paths):
            return func(self, paths=first_arg, **kwargs)

        if first_arg is None:
            first_arg = kwargs.pop("phase_name", None)

        remove_files = conf.instance.general.get("output", "remove_files", bool)

        func(
            self,
            paths=Paths(
                phase_name=first_arg,
                phase_tag=kwargs.pop("phase_tag", None),
                phase_folders=kwargs.pop("phase_folders", tuple()),
                phase_path=kwargs.pop("phase_path", None),
                remove_files=remove_files,
            ),
            **kwargs,
        )

    return wrapper


class Paths:
    def __init__(
        self,
        phase_name="",
        phase_tag=None,
        phase_folders=tuple(),
        phase_path=None,
        remove_files=True,
    ):
        if not isinstance(phase_name, str):
            raise ValueError("Phase name must be a string")
        self.phase_path = phase_path or "/".join(phase_folders)
        self.phase_name = phase_name
        self.phase_tag = phase_tag or ""
        self.remove_files = remove_files

    @property
    def path(self):
        return link.make_linked_folder(self.sym_path)

    def __eq__(self, other):
        return isinstance(other, Paths) and all(
            [
                self.phase_path == other.phase_path,
                self.phase_name == other.phase_name,
                self.phase_tag == other.phase_tag,
            ]
        )

    @property
    def phase_folders(self):
        return self.phase_path.split("/")

    @property
    def backup_path(self) -> str:
        """
        The path to the backed up optimizer folder.
        """
        return f"{self.phase_output_path}/optimizer_backup"

    @property
    def zip_path(self) -> str:
        return f"{self.phase_output_path}.zip"

    @property
    @make_path
    def phase_output_path(self) -> str:
        """
        The path to the output information for a phase.
        """
        return "/".join(
            filter(
                len,
                [
                    conf.instance.output_path,
                    self.phase_path,
                    self.phase_name,
                    self.phase_tag,
                ],
            )
        )

    @property
    def execution_time_path(self) -> str:
        """
        The path to the output information for a phase.
        """
        return "{}/execution_time".format(self.phase_name_folder)

    @property
    @make_path
    def phase_name_folder(self):
        return "/".join((conf.instance.output_path, self.phase_path, self.phase_name))

    @property
    def sym_path(self) -> str:
        return "{}/{}/{}/{}/optimizer".format(
            conf.instance.output_path, self.phase_path, self.phase_name, self.phase_tag
        )

    @property
    def file_param_names(self) -> str:
        return "{}/{}".format(self.path, "multinest.paramnames")

    @property
    def file_model_info(self) -> str:
        return "{}/{}".format(self.phase_output_path, "model.info")

    @property
    @make_path
    def image_path(self) -> str:
        """
        The path to the directory in which images are stored.
        """
        return "{}/image/".format(self.phase_output_path)

    @property
    @make_path
    def pdf_path(self) -> str:
        """
        The path to the directory in which images are stored.
        """
        return "{}pdf/".format(self.image_path)

    def make_optimizer_pickle_path(self) -> str:
        """
        Create the path at which the optimizer pickle should be saved
        """
        return "{}/optimizer.pickle".format(self.make_path())

    def make_model_pickle_path(self):
        """
        Create the path at which the model pickle should be saved
        """
        return "{}/model.pickle".format(self.make_path())

    @make_path
    def make_path(self) -> str:
        """
        Create the path to the folder at which the metadata and optimizer pickle should
        be saved
        """
        return "{}/{}/{}/{}/".format(
            conf.instance.output_path, self.phase_path, self.phase_name, self.phase_tag
        )

    @property
    def file_summary(self) -> str:
        return "{}/{}".format(self.backup_path, "multinestsummary.txt")

    @property
    def file_weighted_samples(self):
        return "{}/{}".format(self.backup_path, "multinest.txt")

    @property
    def file_phys_live(self) -> str:
        return "{}/{}".format(self.backup_path, "multinestphys_live.points")

    @property
    def file_results(self):
        return "{}/{}".format(self.phase_output_path, "model.results")

    def backup(self):
        """
        Copy files from the sym-linked optimizer folder to the backup folder in the workspace.
        """
        try:
            shutil.rmtree(self.backup_path)
        except FileNotFoundError:
            pass

        try:
            shutil.copytree(self.sym_path, self.backup_path)
        except shutil.Error as e:
            logger.exception(e)

    def backup_zip_remove(self):
        """
        Copy files from the sym linked optimizer folder then remove the sym linked folder.
        """
        self.backup()
        self.zip()

        if self.remove_files:
            try:
                shutil.rmtree(self.path)
            except FileNotFoundError:
                pass

    def restore(self):
        """
        Copy files from the backup folder to the sym-linked optimizer folder.
        """
        if os.path.exists(self.zip_path):
            with zipfile.ZipFile(self.zip_path, "r") as f:
                f.extractall(self.phase_output_path)

            os.remove(self.zip_path)

        if os.path.exists(self.backup_path):
            for file in glob.glob(self.backup_path + "/*"):
                shutil.copy(file, self.path)

    def zip(self):
        try:
            with zipfile.ZipFile(self.zip_path, "w", zipfile.ZIP_DEFLATED) as f:
                for root, dirs, files in os.walk(self.phase_output_path):
                    for file in files:
                        f.write(
                            os.path.join(root, file),
                            os.path.join(
                                root[len(self.phase_output_path) :].lstrip("/"), file
                            ),
                        )

            if self.remove_files:
                shutil.rmtree(self.phase_output_path)

        except FileNotFoundError:
            pass



import os

import pytest

import autofit as af
from autofit import Paths
from test_autofit.mock import MockClassNLOx4, MockAnalysis

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="downhill_simplex")
def make_downhill_simplex():
    def fmin(fitness_function, x0):
        fitness_function(x0)
        return x0

    return af.DownhillSimplex(
        fmin=fmin,
        paths=Paths(phase_name="name", phase_folders=("folders",), phase_tag="tag"),
    )


class TestDownhillSimplex:
    def test_instance(self, downhill_simplex, model):
        model.mock_class = MockClassNLOx4()

        assert hasattr(model.instance_from_unit_vector([]), "mock_class")

        result = downhill_simplex.fit(MockAnalysis(), model)

        assert result.instance.mock_class.one == 1
        assert result.instance.mock_class.two == 2
        assert result.figure_of_merit == 1

    def test_model(self, downhill_simplex, model):
        model.mock_class = af.PriorModel(MockClassNLOx4)
        result = downhill_simplex.fit(MockAnalysis(), model)

        assert result.instance.mock_class.one == 0.0
        assert result.instance.mock_class.two == 0.0
        assert result.figure_of_merit == 1

        assert result.model.mock_class.one.mean == 0.0
        assert result.model.mock_class.two.mean == 0.0

    def test_instance_and_model(self, downhill_simplex, model):
        model.instance = MockClassNLOx4()
        model.model = af.PriorModel(MockClassNLOx4)

        result = downhill_simplex.fit(MockAnalysis(), model)

        assert result.instance.instance.one == 1
        assert result.instance.instance.two == 2
        assert result.instance.model.one == 0.0
        assert result.instance.model.two == 0.0
        assert result.model.model.one.mean == 0.0
        assert result.model.model.two.mean == 0.0
        assert result.figure_of_merit == 1


class TestCopyWithNameExtension:
    @staticmethod
    def assert_non_linear_attributes_equal(copy, optimizer):
        assert copy.paths.phase_name == "phase_name/one"

    def test_downhill_simplex(self):
        optimizer = af.DownhillSimplex(Paths("phase_name"), fmin=lambda x: x)

        copy = optimizer.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy, optimizer)
        assert isinstance(copy, af.DownhillSimplex)
        assert copy.fmin is optimizer.fmin
        assert copy.xtol is optimizer.xtol
        assert copy.ftol is optimizer.ftol
        assert copy.maxiter is optimizer.maxiter
        assert copy.maxfun is optimizer.maxfun
        assert copy.full_output is optimizer.full_output
        assert copy.disp is optimizer.disp
        assert copy.retall is optimizer.retall



"""
This package implements the No-U-Turn Sampler (NUTS) algorithm 6 from the NUTS
paper (Hoffman & Gelman, 2011).

Content
-------

The package mainly contains:
  nuts6                     return samples using the NUTS
  numerical_grad            return numerical estimate of the local gradient
  emcee                     emcee NUTS sampler
"""

from .nuts import nuts6
from .helpers import numerical_grad, NutsSampler_fn_wrapper

try:
    from .emcee_nuts import NUTSSampler
except ImportError:
    pass



import logging
import math

import numpy as np
import pymultinest

from autofit import conf, exc
from autofit.optimize.non_linear.non_linear import NonLinearOptimizer
from autofit.optimize.non_linear.non_linear import Result
from autofit.optimize.non_linear.output import NestedSamplingOutput

logger = logging.getLogger(__name__)


class MultiNest(NonLinearOptimizer):
    def __init__(self, paths, sigma_limit=3, run=pymultinest.run):
        """
        Class to setup and run a MultiNest lens and output the MultiNest nlo.

        This interfaces with an input model_mapper, which is used for setting up the \
        individual model instances that are passed to each iteration of MultiNest.
        """

        super().__init__(paths)

        self.sigma_limit = sigma_limit

        self.importance_nested_sampling = self.config(
            "importance_nested_sampling", bool
        )
        self.multimodal = self.config("multimodal", bool)
        self.const_efficiency_mode = self.config("const_efficiency_mode", bool)
        self.n_live_points = self.config("n_live_points", int)
        self.evidence_tolerance = self.config("evidence_tolerance", float)
        self.sampling_efficiency = self.config("sampling_efficiency", float)
        self.n_iter_before_update = self.config("n_iter_before_update", int)
        self.null_log_evidence = self.config("null_log_evidence", float)
        self.max_modes = self.config("max_modes", int)
        self.mode_tolerance = self.config("mode_tolerance", float)
        self.outputfiles_basename = self.config("outputfiles_basename", str)
        self.seed = self.config("seed", int)
        self.verbose = self.config("verbose", bool)
        self.resume = self.config("resume", bool)
        self.context = self.config("context", int)
        self.write_output = self.config("write_output", bool)
        self.log_zero = self.config("log_zero", float)
        self.max_iter = self.config("max_iter", int)
        self.init_MPI = self.config("init_MPI", bool)
        self.run = run

        logger.debug("Creating MultiNest NLO")

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.sigma_limit = self.sigma_limit
        copy.run = self.run
        copy.importance_nested_sampling = self.importance_nested_sampling
        copy.multimodal = self.multimodal
        copy.const_efficiency_mode = self.const_efficiency_mode
        copy.n_live_points = self.n_live_points
        copy.evidence_tolerance = self.evidence_tolerance
        copy.sampling_efficiency = self.sampling_efficiency
        copy.n_iter_before_update = self.n_iter_before_update
        copy.null_log_evidence = self.null_log_evidence
        copy.max_modes = self.max_modes
        copy.mode_tolerance = self.mode_tolerance
        copy.outputfiles_basename = self.outputfiles_basename
        copy.seed = self.seed
        copy.verbose = self.verbose
        copy.resume = self.resume
        copy.context = self.context
        copy.write_output = self.write_output
        copy.log_zero = self.log_zero
        copy.max_iter = self.max_iter
        copy.init_MPI = self.init_MPI
        return copy

    class Fitness(NonLinearOptimizer.Fitness):
        def __init__(
            self, paths, analysis, instance_from_physical_vector, output_results
        ):
            super().__init__(paths, analysis, output_results)
            self.instance_from_physical_vector = instance_from_physical_vector
            self.accepted_samples = 0

            self.model_results_output_interval = conf.instance.general.get(
                "output", "model_results_output_interval", int
            )
            self.stagger_resampling_likelihood = conf.instance.non_linear.get(
                "MultiNest", "stagger_resampling_likelihood", bool
            )
            self.stagger_resampling_value = conf.instance.non_linear.get(
                "MultiNest", "stagger_resampling_value", float
            )
            self.resampling_likelihood = conf.instance.non_linear.get(
                "MultiNest", "null_log_evidence", float
            )
            self.stagger_accepted_samples = 0

        def __call__(self, cube, ndim, nparams, lnew):

            try:
                instance = self.instance_from_physical_vector(cube)
                likelihood = self.fit_instance(instance)
            except exc.FitException:

                if not self.stagger_resampling_likelihood:
                    likelihood = -np.inf
                else:

                    if self.stagger_accepted_samples < 10:

                        self.stagger_accepted_samples += 1
                        self.resampling_likelihood += self.stagger_resampling_value
                        likelihood = self.resampling_likelihood

                    else:

                        likelihood = -1.0 * np.abs(self.resampling_likelihood) * 10.0

            return likelihood

    def fit(self, analysis, model):
        multinest_output = MultiNestOutput(model, self.paths)

        multinest_output.save_model_info()

        # noinspection PyUnusedLocal
        def prior(cube, ndim, nparams):
            # NEVER EVER REFACTOR THIS LINE! Haha.

            phys_cube = model.physical_vector_from_hypercube_vector(
                hypercube_vector=cube
            )

            for i in range(len(phys_cube)):
                cube[i] = phys_cube[i]

            return cube

        fitness_function = MultiNest.Fitness(
            self.paths,
            analysis,
            model.instance_from_physical_vector,
            multinest_output.output_results,
        )

        logger.info("Running MultiNest...")
        self.run(
            fitness_function.__call__,
            prior,
            model.prior_count,
            outputfiles_basename="{}/multinest".format(self.paths.path),
            n_live_points=self.n_live_points,
            const_efficiency_mode=self.const_efficiency_mode,
            importance_nested_sampling=self.importance_nested_sampling,
            evidence_tolerance=self.evidence_tolerance,
            sampling_efficiency=self.sampling_efficiency,
            null_log_evidence=self.null_log_evidence,
            n_iter_before_update=self.n_iter_before_update,
            multimodal=self.multimodal,
            max_modes=self.max_modes,
            mode_tolerance=self.mode_tolerance,
            seed=self.seed,
            verbose=self.verbose,
            resume=self.resume,
            context=self.context,
            write_output=self.write_output,
            log_zero=self.log_zero,
            max_iter=self.max_iter,
            init_MPI=self.init_MPI,
        )
        logger.info("MultiNest complete")

        # TODO: Some of the results below use the backup_path, which isnt updated until the end if thiss function is
        # TODO: not located here. Do we need to rely just ono the optimizer foldeR? This is a good idea if we always
        # TODO: have a valid sym-link( e.g. even for aggregator use).

        self.paths.backup()
        instance = multinest_output.most_likely_model_instance
        analysis.visualize(instance=instance, during_analysis=False)
        multinest_output.output_results(during_analysis=False)
        multinest_output.output_pdf_plots()
        result = Result(
            instance=instance,
            figure_of_merit=multinest_output.evidence,
            previous_model=model,
            gaussian_tuples=multinest_output.gaussian_priors_at_sigma_limit(
                self.sigma_limit
            ),
        )
        self.paths.backup_zip_remove()
        return result

    def output_from_model(self, model, paths):
        return MultiNestOutput(model=model, paths=paths)


class MultiNestOutput(NestedSamplingOutput):
    @property
    def pdf(self):
        import getdist

        try:
            return getdist.mcsamples.loadMCSamples(
                self.paths.backup_path + "/multinest"
            )
        except IOError or OSError or ValueError or IndexError:
            raise Exception

    @property
    def pdf_converged(self):
        try:
            densities_1d = list(
                map(lambda p: self.pdf.get1DDensity(p), self.pdf.getParamNames().names)
            )

            if densities_1d == []:
                return False

            return True
        except Exception:
            return False

    @property
    def most_probable_model_parameters(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt' file which nlo from a \
        multinest lens.

        This file stores the parameters of the most probable model in the first half of entries and the most likely
        model in the second half of entries. The offset parameter is used to start at the desired model.

        """
        try:
            return self.read_list_of_results_from_summary_file(
                number_entries=self.model.prior_count, offset=0
            )
        except FileNotFoundError:
            return self.most_likely_model_parameters

    @property
    def most_likely_model_parameters(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt' file which nlo from a \
        multinest lens.

        This file stores the parameters of the most probable model in the first half of entries and the most likely
        model in the second half of entries. The offset parameter is used to start at the desired model.
        """
        try:
            return self.read_list_of_results_from_summary_file(
                number_entries=self.model.prior_count, offset=56
            )
        except FileNotFoundError:
            most_likey_index = np.argmax([point[-1] for point in self.phys_live_points])
            return self.phys_live_points[most_likey_index][0:-1]

    @property
    def maximum_log_likelihood(self):
        try:
            return self.read_list_of_results_from_summary_file(
                number_entries=2, offset=112
            )[1]
        except FileNotFoundError:
            return max([point[-1] for point in self.phys_live_points])

    @property
    def evidence(self):
        try:
            return self.read_list_of_results_from_summary_file(
                number_entries=2, offset=112
            )[0]
        except FileNotFoundError:
            return None

    @property
    def phys_live_points(self):

        phys_live = open(self.paths.file_phys_live)

        live_points = 0
        for line in phys_live:
            live_points += 1

        phys_live.seek(0)

        phys_live_points = []

        for line in range(live_points):
            vector = []
            for param in range(self.model.prior_count + 1):
                vector.append(float(phys_live.read(28)))
            phys_live.readline()
            phys_live_points.append(vector)

        phys_live.close()

        return phys_live_points

    def phys_live_points_of_param(self, param_index):
        return [point[param_index] for point in self.phys_live_points]

    def read_list_of_results_from_summary_file(self, number_entries, offset):

        summary = open(self.paths.file_summary)
        summary.read(2 + offset * self.model.prior_count)
        vector = []
        for param in range(number_entries):
            vector.append(float(summary.read(28)))

        summary.close()

        return vector

    def model_parameters_at_sigma_limit(self, sigma_limit):
        limit = math.erf(0.5 * sigma_limit * math.sqrt(2))

        if self.pdf_converged:
            densities_1d = list(
                map(lambda p: self.pdf.get1DDensity(p), self.pdf.getParamNames().names)
            )

            return list(map(lambda p: p.getLimits(limit), densities_1d))
        else:

            parameters_min = [
                min(self.phys_live_points_of_param(param_index=param_index))
                for param_index in range(self.model.prior_count)
            ]
            parameters_max = [
                max(self.phys_live_points_of_param(param_index=param_index))
                for param_index in range(self.model.prior_count)
            ]

            return [
                (parameters_min[index], parameters_max[index])
                for index in range(len(parameters_min))
            ]

    @property
    def total_samples(self):
        return len(self.pdf.weights)

    def sample_model_parameters_from_sample_index(self, sample_index):
        """From a sample return the model parameters.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return list(self.pdf.samples[sample_index])

    def sample_weight_from_sample_index(self, sample_index):
        """From a sample return the sample weight.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return self.pdf.weights[sample_index]

    def sample_likelihood_from_sample_index(self, sample_index):
        """From a sample return the likelihood.

        NOTE: GetDist reads the log likelihood from the weighted_sample.txt file (column 2), which are defined as \
        -2.0*likelihood. This routine converts these back to likelihood.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return -0.5 * self.pdf.loglikes[sample_index]



import copy
import inspect
from typing import Tuple, Optional
from functools import wraps

import numpy as np

import autofit.mapper.model
import autofit.mapper.model_mapper
import autofit.mapper.prior_model.collection
from autofit import cast_collection, PriorNameValue, InstanceNameValue
from autofit import exc
from autofit.mapper.model import AbstractModel
from autofit.mapper.prior_model import dimension_type as dim
from autofit.mapper.prior_model.assertion import AbstractAssertion
from autofit.mapper.prior_model.attribute_pair import DeferredNameValue
from autofit.mapper.prior_model.deferred import DeferredArgument
from autofit.mapper.prior_model.prior import GaussianPrior
from autofit.mapper.prior_model.prior import TuplePrior, Prior, WidthModifier
from autofit.mapper.prior_model.recursion import DynamicRecursionCache
from autofit.mapper.prior_model.util import PriorModelNameValue
from autofit.tools.text_formatter import TextFormatter

def check_assertions(func):
    @wraps(func)
    def wrapper(s, arguments):
        # noinspection PyProtectedMember
        for assertion in s._assertions:
            assertion(
                arguments
            )
        return func(s, arguments)

    return wrapper

class AbstractPriorModel(AbstractModel):
    """
    Abstract model that maps a set of priors to a particular class. Must be
    overridden by any prior model so that the model mapper recognises its prior \
    model attributes.

    @DynamicAttrs
    """

    def __init__(self):
        super().__init__()
        self._assertions = list()

    def add_assertion(self, assertion: AbstractAssertion, name=None):
        """
        Assert that some relationship holds between physical values associated with
        priors at the point an instance is created. If this fails a FitException is
        raised causing the model to be re-sampled.

        Parameters
        ----------
        assertion
            An assertion that one prior must be greater than another.
        name
            A name describing the assertion that is logged when it is violated.
        """
        assertion.name = name
        self._assertions.append(assertion)

    @property
    def name(self):
        return self.__class__.__name__

    # noinspection PyUnusedLocal
    @staticmethod
    def from_object(t, *args, **kwargs):
        if inspect.isclass(t):
            from .prior_model import PriorModel

            obj = object.__new__(PriorModel)
            obj.__init__(t, **kwargs)
        elif isinstance(t, list) or isinstance(t, dict):
            obj = object.__new__(
                autofit.mapper.prior_model.collection.CollectionPriorModel
            )
            obj.__init__(t)
        else:
            obj = t
        return obj

    def instance_from_unit_vector(self, unit_vector):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding
        to every PriorModel attributed to this instance.

        This method takes as input a unit vector of parameter values, converting each to
        physical values via their priors.

        Parameters
        ----------
        unit_vector: [float]
            A vector of physical parameter values.

        Returns
        -------
        model_instance : autofit.mapper.model.ModelInstance
            An object containing reconstructed model_mapper instances

        """
        arguments = dict(
            map(
                lambda prior_tuple, unit: (
                    prior_tuple.prior,
                    prior_tuple.prior.value_for(unit),
                ),
                self.prior_tuples_ordered_by_id,
                unit_vector,
            )
        )

        for assertion in self._assertions:
            assertion(arguments)

        return self.instance_for_arguments(arguments)

    @property
    @cast_collection(PriorNameValue)
    def unique_prior_tuples(self):
        """
        Returns
        -------
        prior_tuple_dict: [(Prior, PriorTuple)]
            The set of all priors associated with this mapper
        """
        return {
            prior_tuple[1]: prior_tuple
            for prior_tuple in self.attribute_tuples_with_type(Prior)
        }.values()

    @property
    def unique_promise_tuples(self):
        from autofit import Promise

        return {
            prior_tuple[1]: prior_tuple
            for prior_tuple in self.attribute_tuples_with_type(Promise)
        }.values()

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples_ordered_by_id(self):
        """
        Returns
        -------
        priors: [Prior]
            An ordered list of unique priors associated with this mapper
        """
        return sorted(
            list(self.unique_prior_tuples), key=lambda prior_tuple: prior_tuple.prior.id
        )

    def physical_vector_from_hypercube_vector(self, hypercube_vector):
        """
        Parameters
        ----------
        hypercube_vector: [float]
            A unit hypercube vector

        Returns
        -------
        values: [float]
            A vector with values output by priors
        """
        return list(
            map(
                lambda prior_tuple, unit: prior_tuple.prior.value_for(unit),
                self.prior_tuples_ordered_by_id,
                hypercube_vector,
            )
        )

    @property
    def random_physical_vector_from_priors(self):
        """
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """

        while True:

            physical_vector = self.physical_vector_from_hypercube_vector(
                list(np.random.random(self.prior_count))
            )

            try:
                self.instance_from_physical_vector(physical_vector=physical_vector)
                return physical_vector
            except exc.PriorLimitException:
                pass

    @property
    def physical_values_from_prior_medians(self):
        """
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """
        return self.physical_vector_from_hypercube_vector(
            [0.5] * len(self.unique_prior_tuples)
        )

    def instance_from_physical_vector(self, physical_vector):
        """
        Creates a ModelInstance, which has an attribute and class instance corresponding
        to every PriorModel attributed to this instance.

        This method takes as input a physical vector of parameter values, thus omitting
        the use of priors.

        Parameters
        ----------
        physical_vector: [float]
            A unit hypercube vector

        Returns
        -------
        model_instance : autofit.mapper.model.ModelInstance
            An object containing reconstructed model_mapper instances

        """
        arguments = dict(
            map(
                lambda prior_tuple, physical_unit: (prior_tuple.prior, physical_unit),
                self.prior_tuples_ordered_by_id,
                physical_vector,
            )
        )

        return self.instance_for_arguments(arguments)

    def mapper_from_partial_prior_arguments(self, arguments):
        """
        Creates a new model mapper from a dictionary mapping_matrix existing priors to
        new priors, keeping existing priors where no mapping is provided.

        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping_matrix priors to priors

        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        original_prior_dict = {prior: prior for prior in self.priors}
        return self.mapper_from_prior_arguments({**original_prior_dict, **arguments})

    def mapper_from_prior_arguments(self, arguments):
        """
        Creates a new model mapper from a dictionary mapping_matrix existing priors to
        new priors.

        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping_matrix priors to priors

        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        mapper = copy.deepcopy(self)

        for prior_model_tuple in self.prior_model_tuples:
            setattr(
                mapper,
                prior_model_tuple.name,
                prior_model_tuple.prior_model.gaussian_prior_model_for_arguments(
                    arguments
                ),
            )

        return mapper

    def mapper_from_gaussian_tuples(self, tuples, a=None, r=None):
        """
        Creates a new model mapper from a list of floats describing the mean values
        of gaussian priors. The widths of the new priors are taken from the
        width_config. The new gaussian priors must be provided in the same order as
        the priors associated with model.

        If a is not None then all priors are created with an absolute width of a.

        If r is not None then all priors are created with a relative width of r.

        Parameters
        ----------
        r
            The relative width to be assigned to gaussian priors
        a
            The absolute width to be assigned to gaussian priors
        tuples
            A list of tuples each containing the mean and width of a prior

        Returns
        -------
        mapper: ModelMapper
            A new model mapper with all priors replaced by gaussian priors.
        """

        prior_tuples = self.prior_tuples_ordered_by_id
        prior_class_dict = self.prior_class_dict
        arguments = {}

        for i, prior_tuple in enumerate(prior_tuples):
            prior = prior_tuple.prior
            cls = prior_class_dict[prior]
            mean, sigma = tuples[i]

            def get_name():
                name = prior_tuple.name
                # Use the name of the collection for configuration when a prior's name
                # is just a number (i.e. its position in a collection)
                if name.isdigit():
                    name = self.path_for_prior(prior_tuple.prior)[-2]
                return name

            width_modifier = WidthModifier.for_class_and_attribute_name(cls, get_name())

            if a is not None and r is not None:
                raise exc.PriorException(
                    "Width of new priors cannot be both relative and absolute."
                )
            if a is not None:
                width = a
            elif r is not None:
                width = r * mean
            else:
                width = width_modifier(mean)

            arguments[prior] = GaussianPrior(
                mean, max(tuples[i][1], width), *prior.limits
            )

        return self.mapper_from_prior_arguments(arguments)

    def instance_from_prior_medians(self):
        """
        Creates a list of physical values from the median values of the priors.

        Returns
        -------
        physical_values : [float]
            A list of physical values

        """
        return self.instance_from_unit_vector(
            unit_vector=[0.5] * len(self.prior_tuples)
        )

    @staticmethod
    @DynamicRecursionCache()
    def from_instance(instance, model_classes=tuple()):
        """
        Recursively create an prior object model from an object model.

        Parameters
        ----------
        model_classes
        instance
            A dictionary, list, class instance or model instance

        Returns
        -------
        abstract_prior_model
            A concrete child of an abstract prior model
        """

        if isinstance(instance, list):
            result = autofit.mapper.prior_model.collection.CollectionPriorModel(
                [
                    AbstractPriorModel.from_instance(item, model_classes=model_classes)
                    for item in instance
                ]
            )
        elif isinstance(instance, autofit.mapper.model.ModelInstance):
            result = autofit.mapper.model_mapper.ModelMapper()
            for key, value in instance.dict.items():
                setattr(
                    result,
                    key,
                    AbstractPriorModel.from_instance(
                        value, model_classes=model_classes
                    ),
                )
        elif isinstance(instance, dict):
            result = autofit.mapper.prior_model.collection.CollectionPriorModel(
                {
                    key: AbstractPriorModel.from_instance(
                        value, model_classes=model_classes
                    )
                    for key, value in instance.items()
                }
            )
        elif isinstance(instance, (dim.DimensionType, np.ndarray)):
            return instance
        else:
            from .prior_model import PriorModel

            try:
                result = PriorModel(
                    instance.__class__,
                    **{
                        key: AbstractPriorModel.from_instance(
                            value, model_classes=model_classes
                        )
                        for key, value in instance.__dict__.items()
                        if key != "cls"
                    },
                )
            except AttributeError:
                return instance
        if any([isinstance(instance, cls) for cls in model_classes]):
            return result.as_model()
        return result

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        return self.direct_tuples_with_type(Prior)

    @property
    @cast_collection(InstanceNameValue)
    def direct_instance_tuples(self):
        return self.direct_tuples_with_type(float)

    @property
    @cast_collection(PriorModelNameValue)
    def prior_model_tuples(self):
        return self.direct_tuples_with_type(AbstractPriorModel)

    @property
    @cast_collection(PriorNameValue)
    def tuple_prior_tuples(self):
        """
        Returns
        -------
        tuple_prior_tuples: [(String, TuplePrior)]
        """
        return self.direct_tuples_with_type(TuplePrior)

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        """
        Returns
        -------
        direct_priors: [(String, Prior)]
        """
        return self.direct_tuples_with_type(Prior)

    @property
    @cast_collection(DeferredNameValue)
    def direct_deferred_tuples(self):
        return self.direct_tuples_with_type(DeferredArgument)

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Prior))]
        """
        # noinspection PyUnresolvedReferences
        return self.attribute_tuples_with_type(Prior)

    @property
    @cast_collection(PriorModelNameValue)
    def direct_prior_model_tuples(self):
        return self.direct_tuples_with_type(AbstractPriorModel)

    def __eq__(self, other):
        return (
            isinstance(other, AbstractPriorModel)
            and self.direct_prior_model_tuples == other.direct_prior_model_tuples
        )

    @property
    @cast_collection(InstanceNameValue)
    def instance_tuples(self):
        """
        Returns
        -------
        instances: [(String, instance)]
        """
        return self.attribute_tuples_with_type(float, ignore_class=Prior)

    @property
    def prior_class_dict(self):
        raise NotImplementedError()

    def instance_for_arguments(self, arguments):
        raise NotImplementedError()

    @property
    def prior_count(self):
        return len(self.unique_prior_tuples)

    @property
    def promise_count(self):
        return len(self.unique_promise_tuples)

    @property
    def priors(self):
        return [prior_tuple.prior for prior_tuple in self.prior_tuples]

    def name_for_prior(self, prior):
        for prior_model_name, prior_model in self.direct_prior_model_tuples:
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(prior_model_name, prior_name)
        prior_tuples = self.prior_tuples
        for name, p in prior_tuples:
            if p == prior:
                return name

    def __hash__(self):
        return self.id

    def __add__(self, other):
        result = copy.deepcopy(self)

        for key, value in other.__dict__.items():
            if not hasattr(result, key) or isinstance(value, Prior):
                setattr(result, key, value)
                continue
            result_value = getattr(result, key)
            if isinstance(value, AbstractPriorModel):
                if isinstance(result_value, AbstractPriorModel):
                    setattr(result, key, result_value + value)
                else:
                    setattr(result, key, value)

        return result

    def copy_with_fixed_priors(self, instance, excluded_classes=tuple()):
        """
        Recursively overwrite priors in the mapper with instance values from the
        instance except where the containing class is the descendant of a listed class.

        Parameters
        ----------
        excluded_classes
            Classes that should be left model
        instance
            The best fit from the previous phase
        """
        mapper = copy.deepcopy(self)
        transfer_classes(instance, mapper, excluded_classes)
        return mapper

    @property
    def path_priors_tuples(self):
        path_priors_tuples = self.path_instance_tuples_for_class(Prior)
        return sorted(path_priors_tuples, key=lambda item: item[1].id)

    def path_for_prior(self, prior: Prior) -> Optional[Tuple[str]]:
        """
        Find a path that points at the given tuple.

        Returns the first path or None if no path is found.

        Parameters
        ----------
        prior
            A prior representing what's known about some dimension of the model.

        Returns
        -------
        A path, a series of attributes that point to one location of the prior.
        """
        for path, path_prior in self.path_priors_tuples:
            if path_prior == prior:
                return path
        return None

    @property
    def path_float_tuples(self):
        return self.path_instance_tuples_for_class(float, ignore_class=Prior)

    @property
    def unique_prior_paths(self):
        unique = {item[1]: item for item in self.path_priors_tuples}.values()
        return [item[0] for item in sorted(unique, key=lambda item: item[1].id)]

    @property
    def prior_prior_model_dict(self):
        """
        Returns
        -------
        prior_prior_model_dict: {Prior: PriorModel}
            A dictionary mapping priors to associated prior models. Each prior will only
            have one prior model; if a prior is shared by two prior models then one of
            those prior models will be in this dictionary.
        """
        return {
            prior: prior_model[1]
            for prior_model in self.prior_model_tuples + [("model", self)]
            for _, prior in prior_model[1].prior_tuples
        }

    @property
    def info(self):
        """
        Use the priors that make up the model_mapper to generate information on each
        parameter of the overall model.

        This information is extracted from each priors *model_info* property.
        """
        formatter = TextFormatter()

        for t in self.path_priors_tuples + self.path_float_tuples:
            formatter.add(t)

        return formatter.text

    @property
    def param_names(self):
        """The param_names vector is a list each parameter's analysis_path, and is used
        for *GetDist* visualization.

        The parameter names are determined from the class instance names of the
        model_mapper. Latex tags are properties of each model class."""

        return [
            self.name_for_prior(prior)
            for prior in sorted(self.priors, key=lambda prior: prior.id)
        ]


def transfer_classes(instance, mapper, model_classes=None):
    """
    Recursively overwrite priors in the mapper with instance values from the
    instance except where the containing class is the descendant of a listed class.

    Parameters
    ----------
    model_classes
        Classes whose descendants should not be overwritten
    instance
        The best fit from the previous phase
    mapper
        The prior model from the previous phase
    """
    from autofit.mapper.prior_model.annotation import AnnotationPriorModel

    model_classes = model_classes or []
    for key, instance_value in instance.__dict__.items():
        try:
            mapper_value = getattr(mapper, key)
            if isinstance(mapper_value, Prior) or isinstance(
                mapper_value, AnnotationPriorModel
            ):
                setattr(mapper, key, instance_value)
                continue
            if not any(isinstance(instance_value, cls) for cls in model_classes):
                try:
                    transfer_classes(instance_value, mapper_value, model_classes)
                except AttributeError:
                    setattr(mapper, key, instance_value)
        except AttributeError:
            pass



from test_autofit.mock import MockDataset


def test_save_and_load():
    dataset = MockDataset()
    dataset.save("/tmp")
    assert dataset.name == MockDataset.load(f"/tmp/{dataset.name}.pickle").name



import pytest

import autofit as af


class MockPhaseOutput:
    def __init__(self, directory, pipeline, phase, dataset):
        self.directory = directory
        self.pipeline = pipeline
        self.phase = phase
        self.dataset = dataset


@pytest.fixture(name="aggregator")
def make_aggregator():
    aggregator = af.Aggregator("")
    aggregator.phases = [
        MockPhaseOutput("directory/number/one", "pipeline1", "phase1", "dataset1"),
        MockPhaseOutput("directory/number/two", "pipeline1", "phase2", "dataset1"),
        MockPhaseOutput("directory/letter/a", "pipeline2", "phase2", "dataset2"),
    ]
    return aggregator


def test_attribute(aggregator):
    assert aggregator.pipeline == ["pipeline1", "pipeline1", "pipeline2"]
    assert aggregator.phase == ["phase1", "phase2", "phase2"]
    assert aggregator.dataset == ["dataset1", "dataset1", "dataset2"]


def test_filter(aggregator):
    result = aggregator.filter(pipeline="pipeline1")
    assert len(result) == 2
    assert result[0].pipeline == "pipeline1"

    result = aggregator.filter(pipeline="pipeline1", phase="phase1")

    assert len(result) == 1
    assert result[0].pipeline == "pipeline1"

    result = aggregator.filter(pipeline="pipeline1").filter(phase="phase1")

    assert len(result) == 1
    assert result[0].pipeline == "pipeline1"


def test_filter_contains(aggregator):
    result = aggregator.filter_contains(pipeline="1")
    assert len(result) == 2
    assert result[0].pipeline == "pipeline1"

    result = aggregator.filter_contains(pipeline="1", phase="1")

    assert len(result) == 1
    assert result[0].pipeline == "pipeline1"

    result = aggregator.filter_contains(pipeline="1").filter(phase="phase1")

    assert len(result) == 1
    assert result[0].pipeline == "pipeline1"


def test_filter_contains_directory(aggregator):
    result = aggregator.filter_contains(directory="number")
    assert len(result) == 2

    result = aggregator.filter_contains(directory="letter")
    assert len(result) == 1


def test_group_by(aggregator):
    result = aggregator.group_by("pipeline")

    assert len(result) == 2
    assert len(result[0]) == 2
    assert len(result[1]) == 1

    result = result.filter(phase="phase2")

    assert len(result) == 2
    assert len(result[0]) == 1
    assert len(result[1]) == 1

    assert result.phase == [["phase2"], ["phase2"]]



from copy import deepcopy

import pytest

import autofit as af
from autofit import Paths
from test_autofit.mapper.model.test_model_mapper import GeometryProfile
from test_autofit.mock import GalaxyModel


@pytest.fixture(name="prior_model")
def make_prior_model():
    return af.PriorModel(GeometryProfile)


class TestCase:
    def test_prior_model(self, prior_model):
        prior_model_copy = deepcopy(prior_model)
        assert prior_model == prior_model_copy

        prior_model_copy.centre_0 = af.UniformPrior()

        assert prior_model != prior_model_copy

    def test_list_prior_model(self, prior_model):
        list_prior_model = af.CollectionPriorModel([prior_model])
        list_prior_model_copy = deepcopy(list_prior_model)
        assert list_prior_model == list_prior_model_copy

        list_prior_model[0].centre_0 = af.UniformPrior()

        assert list_prior_model != list_prior_model_copy

    def test_model_mapper(self, prior_model):
        model_mapper = af.ModelMapper()
        model_mapper.prior_model = prior_model
        model_mapper_copy = deepcopy(model_mapper)

        assert model_mapper == model_mapper_copy

        model_mapper.prior_model.centre_0 = af.UniformPrior()

        assert model_mapper != model_mapper_copy

    def test_non_trivial_equality(self):
        model_mapper = af.ModelMapper()
        model_mapper.galaxy = GalaxyModel(
            light_profile=GeometryProfile, mass_profile=GeometryProfile
        )
        model_mapper_copy = deepcopy(model_mapper)

        assert model_mapper == model_mapper_copy

        model_mapper.galaxy.light_profile.centre_0 = af.UniformPrior()

        assert model_mapper != model_mapper_copy

    def test_model_instance_equality(self):
        model_instance = af.ModelInstance()
        model_instance.profile = GeometryProfile()
        model_instance_copy = deepcopy(model_instance)

        assert model_instance == model_instance_copy

        model_instance.profile.centre = (1.0, 2.0)

        assert model_instance != model_instance_copy

    def test_multinest_equality(self):
        nlo = af.MultiNest(Paths("phase name"))
        nlo_copy = deepcopy(nlo)

        assert nlo == nlo_copy

        nlo.n_live_points += 1

        assert nlo != nlo_copy



import autofit as af
from autofit.optimize.non_linear.mock_nlo import MockNLO, MockAnalysis
from test_autofit import mock


def make_pipeline_1(name, phase_folders, optimizer_class):
    phase = af.Phase(
        phase_name="phase_1",
        phase_folders=phase_folders,
        model=af.PriorModel(mock.Galaxy, redshift=af.GaussianPrior(10.0, 1.0)),
        optimizer_class=optimizer_class,
        analysis_class=MockAnalysis,
    )
    return af.Pipeline(f"{name}_1", phase)


def make_pipeline_2(name, phase_folders, optimizer_class):
    phase = af.Phase(
        phase_name="phase_2",
        phase_folders=phase_folders,
        model=af.PriorModel(mock.Galaxy, redshift=af.last.model.redshift),
        optimizer_class=optimizer_class,
        analysis_class=MockAnalysis,
    )
    return af.Pipeline(f"{name}_2", phase)


def make_pipeline(name, phase_folders=tuple(), optimizer_class=MockNLO):
    return make_pipeline_1(name, phase_folders, optimizer_class) + make_pipeline_2(
        name, phase_folders, optimizer_class
    )


def test_pipeline_composition():
    pipeline = make_pipeline("test")
    results = pipeline.run(mock.MockDataset())
    assert results[0].model.redshift.mean == results[1].model.redshift.mean



""" Some helpers to help the usage of NUTS
This package contains:

  numerical_grad            return numerical estimate of the local gradient
_function_wrapper           hack to make partial functions pickleable
NutsSampler_fn_wrapper      combine provided lnp and grad(lnp) into one function
"""
import numpy as np


def numerical_grad(theta, f, dx=1e-3, order=1):
    """ return numerical estimate of the local gradient

    The gradient is computer by using the Taylor expansion approximation over
    each dimension:
        f(t + dt) = f(t) + h df/dt(t) + h^2/2 d^2f/dt^2 + ...

    The first order gives then:
        df/dt = (f(t + dt) - f(t)) / dt + O(dt)
    Note that we could also compute the backwards version by subtracting dt instead:
        df/dt = (f(t) - f(t -dt)) / dt + O(dt)

    A better approach is to use a 3-step formula where we evaluate the
    derivative on both sides of a chosen point t using the above forward and
    backward two-step formulae and taking the average afterward. We need to use the Taylor expansion to higher order:
        f (t +/- dt) = f (t) +/- dt df/dt + dt ^ 2 / 2  dt^2 f/dt^2 +/- dt ^ 3 d^3 f/dt^3 + O(dt ^ 4)

        df/dt = (f(t + dt) - f(t - dt)) / (2 * dt) + O(dt ^ 3)

    Note: that the error is now of the order of dt ^ 3 instead of dt

    In a same manner we can obtain the next order by using f(t +/- 2 * dt):
        df/dt = (f(t - 2 * dt) - 8 f(t - dt)) + 8 f(t + dt) - f(t + 2 * dt) / (12 * dt) + O(dt ^ 4)

    In the 2nd order, two additional function evaluations are required (per dimension), implying a
    more time-consuming algorithm. However the approximation error is of the order of dt ^ 4


    INPUTS
    ------
    theta: ndarray[float, ndim=1]
        vector value around which estimating the gradient
    f: callable
        function from which estimating the gradient

    KEYWORDS
    --------
    dx: float
        pertubation to apply in each direction during the gradient estimation
    order: int in [1, 2]
        order of the estimates:
            1 uses the central average over 2 points
            2 uses the central average over 4 points

    OUTPUTS
    -------
    df: ndarray[float, ndim=1]
        gradient vector estimated at theta

    COST: the gradient estimation need to evaluates ndim * (2 * order) points (see above)
    CAUTION: if dt is very small, the limited numerical precision can result in big errors.
    """
    ndim = len(theta)
    df = np.empty(ndim, dtype=float)
    if order == 1:
        cst = 0.5 / dx
        for k in range(ndim):
            dt = np.zeros(ndim, dtype=float)
            dt[k] = dx
            df[k] = (f(theta + dt) - f(theta - dt)) * cst
    elif order == 2:
        cst = 1.0 / (12.0 * dx)
        for k in range(ndim):
            dt = np.zeros(ndim, dtype=float)
            dt[k] = dx
            df[k] = cst * (
                f(theta - 2 * dt)
                - 8.0 * f(theta - dt)
                + 8.0 * f(theta + dt)
                - f(theta + 2.0 * dt)
            )
    return df


class _function_wrapper:
    """
    This is a hack to make the likelihood function pickleable when ``args``
    are also included.
    """

    def __init__(self, f, args=(), kwargs={}):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self, x):
        try:
            return self.f(x, *self.args, **self.kwargs)
        except:
            import traceback

            print("NUTS: Exception while calling your likelihood function:")
            print("  params:", x)
            print("  args:", self.args)
            print("  kwargs:", self.kwargs)
            print("  exception:")
            traceback.print_exc()
            raise


class NutsSampler_fn_wrapper:
    """ Create a function-like object that combines provided lnp and grad(lnp)
    functions into one as required by nuts6.

    Both functions are stored as partial function allowing to fix arguments if
    the gradient function is not provided, numerical gradient will be computed

    By default, arguments are assumed identical for the gradient and the
    likelihood. However you can change this behavior using set_xxxx_args.
    (keywords are also supported)

    if verbose property is set, each call will print the log-likelihood value
    and the theta point
    """

    def __init__(self, lnp_func, gradlnp_func=None, *args, **kwargs):
        self.lnp_func = _function_wrapper(lnp_func, args, kwargs)
        if gradlnp_func is not None:
            self.gradlnp_func = _function_wrapper(gradlnp_func, args, kwargs)
        else:
            self.gradlnp_func = _function_wrapper(numerical_grad, (self.lnp_func,))
        self.verbose = False

    def set_lnp_args(self, *args, **kwargs):
        self.lnp_func.args = args
        self.lnp_func.kwargs = kwargs

    def set_gradlnp_args(self, *args, **kwargs):
        self.gradlnp_func.args = args
        self.gradlnp_func.kwargs = kwargs

    def __call__(self, theta):
        r = (self.lnp_func(theta), self.gradlnp_func(theta))
        if self.verbose:
            print(r[0], theta)
        return r



import logging

from autofit import conf, exc
from autofit.optimize.non_linear.non_linear import NonLinearOptimizer
from autofit.optimize.non_linear.nuts_src import emcee_nuts
from autofit.optimize.non_linear.output import AbstractOutput

logger = logging.getLogger(__name__)


class NUTS(NonLinearOptimizer):
    def __init__(self, paths, sigma_limit=3):
        """
        Class to setup and run a MultiNest lens and output the MultiNest nlo.

        This interfaces with an input model_mapper, which is used for setting up the \
        individual model instances that are passed to each iteration of MultiNest.
        """

        super().__init__(paths)

        self.sigma_limit = sigma_limit

        logger.debug("Creating NUTS NLO")

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.sigma_limit = self.sigma_limit
        return copy

    class Fitness(NonLinearOptimizer.Fitness):
        def __init__(
            self, paths, analysis, instance_from_physical_vector, output_results
        ):
            super().__init__(paths, analysis, output_results)
            self.instance_from_physical_vector = instance_from_physical_vector
            self.accepted_samples = 0

            self.model_results_output_interval = conf.instance.general.get(
                "output", "model_results_output_interval", int
            )

        def fit_instance(self, instance):
            likelihood = self.analysis.fit(instance)

            # if likelihood > self.max_likelihood:
            #
            #     self.max_likelihood = likelihood
            #     self.result = Result(instance, likelihood)
            #
            #     if self.should_visualize():
            #         self.analysis.visualize(instance, during_analysis=True)
            #
            #     if self.should_backup():
            #         self.paths.backup()
            #
            #     if self.should_output_model_results():
            #         self.output_results(during_analysis=True)

            return likelihood

        def __call__(self, cube):

            try:
                instance = self.instance_from_physical_vector(cube)
                likelihood = self.fit_instance(instance)

                print()
                print(cube)
                print()
                print(likelihood)

            except exc.FitException:

                likelihood = -1.0e8

            return likelihood

    def fit(self, analysis, model):
        output = AbstractOutput(model=model, paths=self.paths)

        output.save_model_info()

        fitness_function = NUTS.Fitness(
            paths=self.paths,
            analysis=analysis,
            instance_from_physical_vector=model.instance_from_physical_vector,
            output_results=None,
        )

        nuts_sampler = emcee_nuts.NUTSSampler(
            dim=model.prior_count,
            log_prob_fn=fitness_function.__call__,
            gradient_fn=None,
        )

        initial_state = model.physical_values_from_prior_medians

        logger.info("Running NUTS Sampling...")
        samples = nuts_sampler.run_mcmc(
            initial_state=initial_state, steps=5000, steps_burn_in=5000, delta=0.6
        )
        logger.info("NUTS complete")

        print(samples)
        stop

        # TODO: Some of the results below use the backup_path, which isnt updated until the end if thiss function is
        # TODO: not located here. Do we need to rely just ono the optimizer foldeR? This is a good idea if we always
        # TODO: have a valid sym-link( e.g. even for aggregator use).

        self.paths.backup()
        #    instance = output.most_likely_model_instance
        #    analysis.visualize(instance=instance, during_analysis=False)
        #    output.output_results(during_analysis=False)
        #    output.output_pdf_plots()
        #     result = Result(
        #         instance=instance,
        #         figure_of_merit=output.evidence,
        #         previous_model=model,
        #         gaussian_tuples=output.gaussian_priors_at_sigma_limit(
        #             self.sigma_limit
        #         ),
        #     )
        #     self.paths.backup_zip_remove()
        return None



from abc import ABC, abstractmethod

from autofit import exc


class AbstractAssertion(ABC):
    def __init__(self):
        self.name = None

    @abstractmethod
    def __call__(self, arg_dict: dict):
        """

        """


class ComparisonAssertion(AbstractAssertion, ABC):
    def __init__(self, lower, greater):
        """
        Describes an assertion that the physical values associated with
        the lower and greater priors are lower and greater respectively.

        Parameters
        ----------
        lower: Prior
            A prior object with physical values that must be lower
        greater: Prior
            A prior object with physical values that must be greater
        """
        super().__init__()
        self._lower = lower
        self._greater = greater

    def __gt__(self, other):
        return CompoundAssertion(self, self._lower > other)

    def __lt__(self, other):
        return CompoundAssertion(self, self._greater < other)

    def __ge__(self, other):
        return CompoundAssertion(self, self._lower >= other)

    def __le__(self, other):
        return CompoundAssertion(self, self._greater <= other)

    def lower(self, arg_dict: dict):
        if isinstance(self._lower, float):
            return self._lower
        return arg_dict[self._lower]

    def greater(self, arg_dict: dict):
        if isinstance(self._greater, float):
            return self._greater
        return arg_dict[self._greater]


class GreaterThanLessThanAssertion(ComparisonAssertion):
    def __call__(self, arg_dict: dict):
        """
        Assert that the value in the dictionary associated with the lower
        prior is lower than the value associated with the greater prior.

        Parameters
        ----------
        arg_dict
            A dictionary mapping priors to physical values.

        Raises
        ------
        FitException
            If the assertion is not met
        """
        if self.lower(arg_dict) > self.greater(arg_dict):
            raise exc.FitException(
                "Assertion failed" + ("" if self.name is None else f" '{self.name}'")
            )


class GreaterThanLessThanEqualAssertion(ComparisonAssertion):
    def __call__(self, arg_dict: dict):
        """
        Assert that the value in the dictionary associated with the lower
        prior is lower than the value associated with the greater prior.

        Parameters
        ----------
        arg_dict
            A dictionary mapping priors to physical values.

        Raises
        ------
        FitException
            If the assertion is not met
        """
        if not self.lower(arg_dict) <= self.greater(arg_dict):
            raise exc.FitException(
                "Assertion failed" + ("" if self.name is None else f" '{self.name}'")
            )


class CompoundAssertion(AbstractAssertion):
    def __init__(self, assertion_1, assertion_2):
        super().__init__()
        self.assertion_1 = assertion_1
        self.assertion_2 = assertion_2

    def __call__(self, arg_dict: dict):
        self.assertion_1(arg_dict)
        self.assertion_2(arg_dict)



import itertools
import os
import shutil

import pytest

import autofit as af
from autofit import ModelMapper, Paths
from autofit.optimize.non_linear.output import AbstractOutput
from test_autofit.mock import (
    GeometryProfile,
    MockClassNLOx4,
    MockClassNLOx5,
    MockNonLinearOptimizer,
    MockClassNLOx6,
)

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="mapper")
def make_mapper():
    return af.ModelMapper()


@pytest.fixture(name="mock_list")
def make_mock_list():
    return [af.PriorModel(MockClassNLOx4), af.PriorModel(MockClassNLOx4)]


@pytest.fixture(name="result")
def make_result():
    mapper = af.ModelMapper()
    mapper.profile = GeometryProfile
    # noinspection PyTypeChecker
    return af.Result(None, None, mapper, [(0, 0), (1, 0)])


class TestResult:
    def test_model(self, result):
        profile = result.model.profile
        assert profile.centre_0.mean == 0
        assert profile.centre_1.mean == 1
        assert profile.centre_0.sigma == 0.05
        assert profile.centre_1.sigma == 0.05

    def test_model_absolute(self, result):
        profile = result.model_absolute(a=2.0).profile
        assert profile.centre_0.mean == 0
        assert profile.centre_1.mean == 1
        assert profile.centre_0.sigma == 2.0
        assert profile.centre_1.sigma == 2.0

    def test_model_relative(self, result):
        profile = result.model_relative(r=1.0).profile
        assert profile.centre_0.mean == 0
        assert profile.centre_1.mean == 1
        assert profile.centre_0.sigma == 0.0
        assert profile.centre_1.sigma == 1.0

    def test_raises(self, result):
        with pytest.raises(af.exc.PriorException):
            result.model.mapper_from_gaussian_tuples(
                result.gaussian_tuples, a=2.0, r=1.0
            )


class TestCopyWithNameExtension:
    @staticmethod
    def assert_non_linear_attributes_equal(copy):
        assert copy.paths.phase_name == "phase_name/one"

    def test_copy_with_name_extension(self):
        optimizer = af.NonLinearOptimizer(Paths("phase_name", phase_tag="tag"))
        copy = optimizer.copy_with_name_extension("one")

        self.assert_non_linear_attributes_equal(copy)
        assert optimizer.paths.phase_tag == copy.paths.phase_tag

    def test_grid_search(self):
        optimizer = af.GridSearch(Paths("phase_name"), step_size=17, grid=lambda x: x)

        copy = optimizer.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy)
        assert isinstance(copy, af.GridSearch)
        assert copy.step_size is optimizer.step_size
        assert copy.grid is optimizer.grid


@pytest.fixture(name="nlo_setup_path")
def test_nlo_setup():
    nlo_setup_path = "{}/../test_files/non_linear/nlo/setup/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    if os.path.exists(nlo_setup_path):
        shutil.rmtree(nlo_setup_path)

    os.mkdir(nlo_setup_path)

    return nlo_setup_path


@pytest.fixture(name="nlo_model_info_path")
def test_nlo_model_info():
    nlo_model_info_path = "{}/../test_files/non_linear/nlo/model_info/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    if os.path.exists(nlo_model_info_path):
        shutil.rmtree(nlo_model_info_path)

    return nlo_model_info_path


@pytest.fixture(name="nlo_wrong_info_path")
def test_nlo_wrong_info():
    nlo_wrong_info_path = "{}/../test_files/non_linear/nlo/wrong_info/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    if os.path.exists(nlo_wrong_info_path):
        shutil.rmtree(nlo_wrong_info_path)

    os.mkdir(nlo_wrong_info_path)

    return nlo_wrong_info_path


class TestDirectorySetup:
    def test__1_class__correct_directory(self, nlo_setup_path):
        af.conf.instance.output_path = nlo_setup_path + "1_class"
        af.NonLinearOptimizer(Paths(phase_name=""))

        assert os.path.exists(nlo_setup_path + "1_class")


class TestMostProbableAndLikely:
    def test__most_probable_parameters_and_instance__2_classes_6_params(self):
        mapper = af.ModelMapper(
            mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6
        )
        nlo = MockNonLinearOptimizer(
            phase_name="",
            model_mapper=mapper,
            most_probable=[1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0],
        )

        most_probable = nlo.most_probable_model_instance

        assert most_probable.mock_class_1.one == 1.0
        assert most_probable.mock_class_1.two == 2.0
        assert most_probable.mock_class_1.three == 3.0
        assert most_probable.mock_class_1.four == 4.0

        assert most_probable.mock_class_2.one == (-5.0, -6.0)
        assert most_probable.mock_class_2.two == (-7.0, -8.0)
        assert most_probable.mock_class_2.three == 9.0
        assert most_probable.mock_class_2.four == 10.0

    def test__most_probable__setup_model_instance__1_class_5_params_but_1_is_instance(
        self
    ):
        mapper = af.ModelMapper(mock_class=MockClassNLOx5)
        mapper.mock_class.five = 10.0

        nlo = MockNonLinearOptimizer(
            phase_name="",
            model_mapper=mapper,
            most_probable=[1.0, -2.0, 3.0, 4.0, 10.0],
        )

        most_probable = nlo.most_probable_model_instance

        assert most_probable.mock_class.one == 1.0
        assert most_probable.mock_class.two == -2.0
        assert most_probable.mock_class.three == 3.0
        assert most_probable.mock_class.four == 4.0
        assert most_probable.mock_class.five == 10.0

    def test__most_likely_parameters_and_instance__2_classes_6_params(self):
        mapper = af.ModelMapper(
            mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6
        )
        nlo = MockNonLinearOptimizer(
            phase_name="",
            model_mapper=mapper,
            most_likely=[21.0, 22.0, 23.0, 24.0, 25.0, -26.0, -27.0, 28.0, 29.0, 30.0],
        )

        most_likely = nlo.most_likely_model_instance

        assert most_likely.mock_class_1.one == 21.0
        assert most_likely.mock_class_1.two == 22.0
        assert most_likely.mock_class_1.three == 23.0
        assert most_likely.mock_class_1.four == 24.0

        assert most_likely.mock_class_2.one == (25.0, -26.0)
        assert most_likely.mock_class_2.two == (-27.0, 28.0)
        assert most_likely.mock_class_2.three == 29.0
        assert most_likely.mock_class_2.four == 30.0

    def test__most_likely__setup_model_instance__1_class_5_params_but_1_is_instance(
        self
    ):
        mapper = af.ModelMapper(mock_class=MockClassNLOx5)
        mapper.mock_class.five = 10.0
        nlo = MockNonLinearOptimizer(
            phase_name="",
            model_mapper=mapper,
            most_likely=[9.0, -10.0, -11.0, 12.0, 10.0],
        )

        most_likely = nlo.most_likely_model_instance

        assert most_likely.mock_class.one == 9.0
        assert most_likely.mock_class.two == -10.0
        assert most_likely.mock_class.three == -11.0
        assert most_likely.mock_class.four == 12.0
        assert most_likely.mock_class.five == 10.0


class TestGaussianPriors:
    def test__1_class__gaussian_priors_at_3_sigma_confidence(self):
        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
        nlo = MockNonLinearOptimizer(
            phase_name="",
            model_mapper=mapper,
            most_probable=[1.0, 2.0, 3.0, 4.1],
            model_lower_params=[0.88, 1.88, 2.88, 3.88],
            model_upper_params=[1.12, 2.12, 3.12, 4.12],
        )

        gaussian_priors = nlo.gaussian_priors_at_sigma_limit(sigma_limit=3.0)

        assert gaussian_priors[0][0] == 1.0
        assert gaussian_priors[1][0] == 2.0
        assert gaussian_priors[2][0] == 3.0
        assert gaussian_priors[3][0] == 4.1

        assert gaussian_priors[0][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[1][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[2][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[3][1] == pytest.approx(0.22, 1e-2)


class TestOffsetFromInput:
    def test__input_model_offset_from_most_probable__parameters_and_instance__1_class_4_params(
        self
    ):
        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
        nlo = MockNonLinearOptimizer(
            phase_name="", model_mapper=mapper, most_probable=[1.0, -2.0, 3.0, 4.0]
        )

        offset_values = nlo.values_offset_from_input_model_parameters(
            input_model_parameters=[1.0, 1.0, 2.0, 3.0]
        )

        assert offset_values == [0.0, -3.0, 1.0, 1.0]

        mapper = af.ModelMapper(
            mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6
        )
        nlo = MockNonLinearOptimizer(
            phase_name="",
            model_mapper=mapper,
            most_probable=[1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0],
        )

        offset_values = nlo.values_offset_from_input_model_parameters(
            input_model_parameters=[
                1.0,
                1.0,
                2.0,
                3.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                20.0,
            ]
        )

        assert offset_values == [
            0.0,
            1.0,
            1.0,
            1.0,
            -15.0,
            -16.0,
            -17.0,
            -18.0,
            -1.0,
            -10.0,
        ]


@pytest.fixture(name="optimizer")
def make_optimizer():
    return AbstractOutput(ModelMapper(), Paths(phase_name=""))


class TestLabels:
    def test_param_names(self, optimizer):
        optimizer.model.prior_model = MockClassNLOx4
        assert [
            "prior_model_one",
            "prior_model_two",
            "prior_model_three",
            "prior_model_four",
        ] == optimizer.model.param_names

    def test_properties(self, optimizer):
        optimizer.model.prior_model = MockClassNLOx4

        assert len(optimizer.param_labels) == 4
        assert len(optimizer.model.param_names) == 4

    def test_label_config(self):
        assert af.conf.instance.label.label("one") == "x4p0"
        assert af.conf.instance.label.label("two") == "x4p1"
        assert af.conf.instance.label.label("three") == "x4p2"
        assert af.conf.instance.label.label("four") == "x4p3"

    def test_labels(self, optimizer):
        af.AbstractPriorModel._ids = itertools.count()
        optimizer.model.prior_model = MockClassNLOx4

        assert optimizer.param_labels == [
            r"x4p0_{\mathrm{a1}}",
            r"x4p1_{\mathrm{a1}}",
            r"x4p2_{\mathrm{a1}}",
            r"x4p3_{\mathrm{a1}}",
        ]



import os
import shutil

import pytest

import autofit.optimize.non_linear.grid_search
import autofit.optimize.non_linear.non_linear
from autofit import conf, Paths
from autofit import exc
from autofit.optimize.optimizer import grid
from test_autofit.mock import Galaxy


class MockAnalysis(autofit.optimize.non_linear.non_linear.Analysis):
    def __init__(self):
        self.instances = []

    def fit(self, instance):
        self.instances.append(instance)
        try:
            return -(
                (instance.one.redshift - 0.1) ** 2 + (instance.two.redshift - 0.7) ** 2
            )
        except AttributeError:
            return 0

    def visualize(self, instance, during_analysis):
        pass


def tuple_lists_equal(l1, l2):
    assert len(l1) == len(l2)
    for tuple_pair in zip(l1, l2):
        assert len(tuple_pair[0]) == len(tuple_pair[1])
        for item in zip(tuple_pair[0], tuple_pair[1]):
            if item[0] != pytest.approx(item[1]):
                return False
    return True


class TestGridSearchOptimizer:
    def test_config(self):
        assert (
            autofit.optimize.non_linear.grid_search.GridSearch(
                Paths(phase_name="")
            ).step_size
            == 0.1
        )

    def test_1d(self):
        points = []

        def fit(point):
            points.append(point)
            return 0

        grid(fit, 1, 0.1)

        assert 10 == len(points)
        assert tuple_lists_equal(
            [
                (0.05,),
                (0.15,),
                (0.25,),
                (0.35,),
                (0.45,),
                (0.55,),
                (0.65,),
                (0.75,),
                (0.85,),
                (0.95,),
            ],
            points,
        )

    def test_2d(self):
        points = []

        def fit(point):
            points.append(point)
            return 0

        grid(fit, 2, 0.3)

        assert 9 == len(points)
        assert tuple_lists_equal(
            [
                (0.15, 0.15),
                (0.15, 0.45),
                (0.15, 0.75),
                (0.45, 0.15),
                (0.45, 0.45),
                (0.45, 0.75),
                (0.75, 0.15),
                (0.75, 0.45),
                (0.75, 0.75),
            ],
            points,
        )

    def test_3d(self):
        points = []

        def fit(point):
            points.append(point)
            return 0

        grid(fit, 3, 0.5)

        assert 3 == len(points[0])
        assert 8 == len(points)

    def test_best_fit(self):
        best_point = (0.65, 0.35)

        def fit(point):
            return -((best_point[0] - point[0]) ** 2 + (best_point[1] - point[1]) ** 2)

        result = grid(fit, 2, 0.1)

        assert pytest.approx(result[0]) == best_point[0]
        assert pytest.approx(result[1]) == best_point[1]


@pytest.fixture(name="grid_search")
def make_grid_search():
    name = "grid_search"
    try:
        shutil.rmtree("{}/{}/".format(conf.instance.output_path, name))
    except FileNotFoundError:
        pass
    return autofit.optimize.non_linear.grid_search.GridSearch(
        Paths(phase_name=name, remove_files=False), step_size=0.1
    )


class TestGridSearch:
    def test_1d(self, grid_search, model):
        model.one = Galaxy

        analysis = MockAnalysis()
        grid_search.fit(analysis, model)

        assert len(analysis.instances) == 10

        instance = analysis.instances[5]

        assert isinstance(instance.one, Galaxy)
        assert instance.one.redshift == 0.55

    def test_2d(self, grid_search, model):
        model.one = Galaxy
        model.two = Galaxy

        analysis = MockAnalysis()

        result = grid_search.fit(analysis, model)

        assert pytest.approx(result.instance.one.redshift) == 0.05
        assert pytest.approx(result.instance.two.redshift) == 0.65

    def test_checkpoint_properties(self, grid_search, model):
        analysis = MockAnalysis()

        model.one = Galaxy
        grid_search.fit(analysis, model)

        paths = Paths(phase_name="grid_search")

        grid_search = autofit.optimize.non_linear.grid_search.GridSearch(
            paths, step_size=0.1
        )

        assert grid_search.is_checkpoint
        assert grid_search.checkpoint_count == 10
        assert grid_search.checkpoint_fit == 0.0
        assert grid_search.checkpoint_cube == (0.05,)
        assert grid_search.checkpoint_step_size == 0.1
        assert grid_search.checkpoint_prior_count == 1

    def test_recover_bad_checkpoint(self, grid_search, model):
        analysis = MockAnalysis()

        model.one = Galaxy
        grid_search.fit(analysis, model)

        grid_search = autofit.optimize.non_linear.grid_search.GridSearch(
            Paths(phase_name="grid_search"), step_size=0.1
        )

        model.two = Galaxy

        with pytest.raises(exc.CheckpointException):
            grid_search.fit(analysis, model)

        grid_search = autofit.optimize.non_linear.grid_search.GridSearch(
            Paths(phase_name="grid_search"), step_size=0.2
        )
        model.one = Galaxy

        with pytest.raises(exc.CheckpointException):
            grid_search.fit(analysis, model)

    def test_recover_checkpoint(self, grid_search, model):
        analysis = MockAnalysis()

        model.one = Galaxy
        model.two = Galaxy

        grid_search.fit(analysis, model)

        grid_search = autofit.optimize.non_linear.grid_search.GridSearch(
            Paths(phase_name="grid_search"), step_size=0.1
        )

        analysis = MockAnalysis()

        result = grid_search.fit(analysis, model)

        assert len(analysis.instances) == 1
        assert pytest.approx(result.instance.one.redshift) == 0.05
        assert pytest.approx(result.instance.two.redshift) == 0.65

    def test_recover_midway(self, grid_search, model):
        string = "11\n0\n(0.0, 0.0)\n0.1\n2"
        with open(grid_search.checkpoint_path, "w+") as f:
            f.write(string)

        grid_search = autofit.optimize.non_linear.grid_search.GridSearch(
            Paths(phase_name="grid_search"), step_size=0.1
        )

        model.one = Galaxy
        model.two = Galaxy

        analysis = MockAnalysis()

        result = grid_search.fit(analysis, model)

        assert len(analysis.instances) == 90
        assert pytest.approx(result.instance.one.redshift) == 0.15
        assert pytest.approx(result.instance.two.redshift) == 0.65

    def test_instances(self, grid_search, model):
        model.one = Galaxy

        analysis = MockAnalysis()
        result = grid_search.fit(analysis, model)

        assert len(result.instances) == 10

        instance = result.instances[5]

        assert isinstance(instance[0].one, Galaxy)
        assert instance[0].one.redshift == 0.55



#!/usr/bin/env python

"""
Filter and collate phase dataset in all subdirectories.

Usage:

./aggregator.py (root_directory) [pipeline=pipeline phase=phase dataset=dataset]

Example:

./aggregator.py ../output pipeline=data_mass_x1_source_x1_positions
"""

import os
import pickle
import zipfile
from collections import defaultdict
from typing import List

import autofit.optimize.non_linear.non_linear
from autofit.optimize.non_linear.output import AbstractOutput


class PhaseOutput:
    """
    @DynamicAttrs
    """

    def __init__(self, directory: str):
        """
        Represents the output of a single phase. Comprises a metadata file and other dataset files.

        Parameters
        ----------
        directory
            The directory of the phase
        """
        self.directory = directory
        self.__optimizer = None
        self.__model = None
        self.file_path = os.path.join(directory, "metadata")
        with open(self.file_path) as f:
            self.text = f.read()
            pairs = [line.split("=") for line in self.text.split("\n")]
            self.__dict__.update({pair[0]: pair[1] for pair in pairs})

    @property
    def output(self) -> AbstractOutput:
        """
        An object describing the output data from the nonlinear search performed in this phase
        """
        return self.optimizer.output_from_model(
            model=self.model, paths=self.optimizer.paths
        )

    @property
    def model_results(self) -> str:
        """
        Reads the model.results file
        """
        with open(os.path.join(self.directory, "model.results")) as f:
            return f.read()

    @property
    def dataset(self):
        """
        The dataset that this phase ran on
        """
        with open(
            os.path.join(self.directory, f"{self.dataset_name}.pickle"), "rb"
        ) as f:
            return pickle.load(f)

    @property
    def header(self) -> str:
        """
        A header created by joining the pipeline, phase and dataset names
        """
        return "/".join((self.pipeline, self.phase, self.dataset_name))

    @property
    def optimizer(self) -> autofit.optimize.non_linear.non_linear.NonLinearOptimizer:
        """
        The optimizer object that was used in this phase
        """
        if self.__optimizer is None:
            with open(os.path.join(self.directory, "optimizer.pickle"), "r+b") as f:
                self.__optimizer = pickle.loads(f.read())
        return self.__optimizer

    @property
    def model(self) -> autofit.optimize.non_linear.non_linear.NonLinearOptimizer:
        """
        The optimizer object that was used in this phase
        """
        if self.__model is None:
            with open(os.path.join(self.directory, "model.pickle"), "r+b") as f:
                self.__model = pickle.loads(f.read())
        return self.__model

    def __str__(self):
        return self.text

    def __repr__(self):
        return "<PhaseOutput {}>".format(self)


class AggregatorGroup:
    def __init__(self, groups: ["AbstractAggregator"]):
        """
        A group of aggregators produced by grouping phases on a field.

        Parameters
        ----------
        groups
            Groups, each with a common value in the metadata file
        """
        self.groups = groups

    def filter(self, **kwargs) -> "AggregatorGroup":
        """
        Apply a filter to the underlying groups whilst maintaining the total number of groups.

        Parameters
        ----------
        kwargs
            Key word arguments for filtering

        Returns
        -------
        A collection of groups of the same length with each group having the same or fewer members.
        """
        return AggregatorGroup([group.filter(**kwargs) for group in self.groups])

    def __getitem__(self, item):
        return self.groups[item]

    def __len__(self):
        return len(self.groups)

    def __getattr__(self, item):
        return [getattr(group, item) for group in self.groups]


class AbstractAggregator:
    def __init__(self, phases: List[PhaseOutput]):
        """
        An aggregator that comprises several phases which matching filters.

        Parameters
        ----------
        phases
            Phases that were found to have matching filters
        """
        self.phases = phases

    def __getitem__(self, item):
        return self.phases[item]

    def __len__(self):
        return len(self.phases)

    def phases_with(self, **kwargs) -> [PhaseOutput]:
        """
        Filters phases. If no arguments are passed all phases are returned. Arguments must be key value pairs, with
        phase, dataset or pipeline as the key.

        Parameters
        ----------
        kwargs
            Filters, e.g. pipeline=pipeline1
        """

        return [
            phase
            for phase in self.phases
            if all([getattr(phase, key) == value for key, value in kwargs.items()])
        ]

    def phases_with_contains(self, **kwargs) -> [PhaseOutput]:
        """
        Filters phases by checking if an attribute contains some substring. If no arguments are passed all phases
        are returned. Arguments must be key value pairs, with phase, dataset or pipeline as the key.

        Parameters
        ----------
        kwargs
            Filters, e.g. pipeline1 in pipeline
        """
        return [
            phase
            for phase in self.phases
            if all([value in getattr(phase, key) for key, value in kwargs.items()])
        ]

    def filter(self, **kwargs) -> "AbstractAggregator":
        """
        Filter by key value pairs found in the metadata.

        Another aggregator object is returned.

        Parameters
        ----------
        kwargs
            Key value arguments which are expected to match those found the in metadata file.

        Returns
        -------
        An aggregator comprising all phases that match the filters.
        """
        return AbstractAggregator(phases=self.phases_with(**kwargs))

    def filter_contains(self, **kwargs) -> "AbstractAggregator":
        """
        Filter by key value pairs found in the metadata. Checks whether the value contains
        the given filter value.

        Another aggregator object is returned.

        Parameters
        ----------
        kwargs
            Key value arguments which are expected to match those found the in metadata file.

        Returns
        -------
        An aggregator comprising all phases that match the filters.
        """
        return AbstractAggregator(phases=self.phases_with_contains(**kwargs))

    def __getattr__(self, item):
        """
        If an attribute is not found then attempt to grab a list of values from the underlying phases
        """
        return [getattr(phase, item) for phase in self.phases]

    def group_by(self, field: str) -> AggregatorGroup:
        """
        Group the phases by a field, e.g. pipeline.

        The object returned still permits filtering and attribute querying.

        Parameters
        ----------
        field
            The field by which to group

        Returns
        -------
        An object comprising lists of grouped fields
        """
        group_dict = defaultdict(list)
        for phase in self.phases:
            group_dict[getattr(phase, field)].append(phase)
        return AggregatorGroup(list(map(AbstractAggregator, group_dict.values())))

    @property
    def model_results(self) -> str:
        """
        A string joining headers and results for all included phases.
        """
        return "\n\n".join(
            "{}\n\n{}".format(phase.header, phase.model_results)
            for phase in self.phases
        )


class Aggregator(AbstractAggregator):
    def __init__(self, directory: str):
        """
        Class to aggregate phase results for all subdirectories in a given directory.

        The whole directory structure is traversed and a Phase object created for each directory that contains a
        metadata file.

        Parameters
        ----------
        directory
        """
        self.directory = directory
        phases = []

        for root, _, filenames in os.walk(directory):
            for filename in filenames:
                if filename.endswith(".zip"):
                    with zipfile.ZipFile(f"{root}/{filename}", "r") as f:
                        f.extractall(f"{root}/{filename[:-4]}")

        for root, _, filenames in os.walk(directory):
            if "metadata" in filenames:
                phases.append(PhaseOutput(root))

        if len(phases) == 0:
            print(f"\nNo phases found in {directory}\n")
        else:
            paths_string = "\n".join(phase.directory for phase in phases)
            print(f"\nPhases were found in these directories:\n\n{paths_string}\n")
        super().__init__(phases)


if __name__ == "__main__":
    from sys import argv

    root_directory = None
    try:
        root_directory = argv[1]
    except IndexError:
        print(
            "Usage:\n\naggregator.py (root_directory) [pipeline=pipeline phase=phase dataset=dataset]"
        )
        exit(1)
    filter_dict = {pair[0]: pair[1] for pair in [line.split("=") for line in argv[2:]]}

    with open("model.results", "w+") as out:
        out.write(Aggregator(root_directory).filter(**filter_dict).model_results)



import logging
import math
import os

import emcee
import numpy as np

from autofit import conf, exc
from autofit.optimize.non_linear.non_linear import NonLinearOptimizer
from autofit.optimize.non_linear.non_linear import Result
from autofit.optimize.non_linear.output import MCMCOutput

logger = logging.getLogger(__name__)


class Emcee(NonLinearOptimizer):
    def __init__(self, paths, sigma_limit=3):
        """
        Class to setup and run a MultiNest lens and output the MultiNest nlo.

        This interfaces with an input model_mapper, which is used for setting up the \
        individual model instances that are passed to each iteration of MultiNest.
        """

        super().__init__(paths)

        self.sigma_limit = sigma_limit

        self.nwalkers = conf.instance.non_linear.get("Emcee", "nwalkers", int)
        self.nsteps = conf.instance.non_linear.get("Emcee", "nsteps", int)
        self.check_auto_correlation = conf.instance.non_linear.get(
            "Emcee", "check_auto_correlation", bool
        )
        self.auto_correlation_check_size = conf.instance.non_linear.get(
            "Emcee", "auto_correlation_check_size", int
        )
        self.auto_correlation_required_length = conf.instance.non_linear.get(
            "Emcee", "auto_correlation_required_length", int
        )
        self.auto_correlation_change_threshold = conf.instance.non_linear.get(
            "Emcee", "auto_correlation_change_threshold", float
        )

        logger.debug("Creating Emcee NLO")

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.sigma_limit = self.sigma_limit
        return copy

    class Fitness(NonLinearOptimizer.Fitness):
        def __init__(
            self, paths, analysis, instance_from_physical_vector, output_results
        ):
            super().__init__(paths, analysis, output_results)
            self.instance_from_physical_vector = instance_from_physical_vector
            self.accepted_samples = 0

        def fit_instance(self, instance):
            likelihood = self.analysis.fit(instance)

            if likelihood > self.max_likelihood:

                self.max_likelihood = likelihood
                self.result = Result(instance, likelihood)

                if self.should_visualize():
                    self.analysis.visualize(instance, during_analysis=True)

                if self.should_backup():
                    self.paths.backup()

                if self.should_output_model_results():
                    self.output_results(during_analysis=True)

            return likelihood

        def __call__(self, params):

            try:

                instance = self.instance_from_physical_vector(params)
                likelihood = self.fit_instance(instance)

            except exc.FitException:

                likelihood = -np.inf

            return likelihood

    def fit(self, analysis, model):

        output = EmceeOutput(
            model=model,
            paths=self.paths,
            auto_correlation_check_size=self.auto_correlation_check_size,
            auto_correlation_required_length=self.auto_correlation_required_length,
            auto_correlation_change_threshold=self.auto_correlation_change_threshold,
        )

        fitness_function = Emcee.Fitness(
            paths=self.paths,
            analysis=analysis,
            instance_from_physical_vector=model.instance_from_physical_vector,
            output_results=output.output_results,
        )

        emcee_sampler = emcee.EnsembleSampler(
            nwalkers=self.nwalkers,
            ndim=model.prior_count,
            log_prob_fn=fitness_function.__call__,
            backend=emcee.backends.HDFBackend(filename=self.paths.path + "/emcee.hdf"),
        )

        output.save_model_info()

        try:
            emcee_state = emcee_sampler.get_last_sample()
            previuos_run_converged = output.converged

        except AttributeError:

            emcee_state = np.zeros(shape=(emcee_sampler.nwalkers, emcee_sampler.ndim))

            for walker_index in range(emcee_sampler.nwalkers):
                emcee_state[walker_index, :] = np.asarray(
                    model.random_physical_vector_from_priors
                )

            previuos_run_converged = False

        logger.info("Running Emcee Sampling...")

        if self.nsteps - emcee_sampler.iteration > 0 and not previuos_run_converged:

            for sample in emcee_sampler.sample(
                initial_state=emcee_state,
                iterations=self.nsteps - emcee_sampler.iteration,
                progress=True,
                skip_initial_state_check=True,
                store=True,
            ):

                if emcee_sampler.iteration % self.auto_correlation_check_size:
                    continue

                if output.converged and self.check_auto_correlation:
                    break

        logger.info("Emcee complete")

        # TODO: Some of the results below use the backup_path, which isnt updated until the end if thiss function is
        # TODO: not located here. Do we need to rely just ono the optimizer foldeR? This is a good idea if we always
        # TODO: have a valid sym-link( e.g. even for aggregator use).

        self.paths.backup()

        instance = output.most_likely_model_instance

        analysis.visualize(instance=instance, during_analysis=False)
        output.output_results(during_analysis=False)
        output.output_pdf_plots()
        result = Result(
            instance=instance,
            figure_of_merit=output.maximum_log_likelihood,
            previous_model=model,
            gaussian_tuples=output.gaussian_priors_at_sigma_limit(self.sigma_limit),
        )
        self.paths.backup_zip_remove()
        return result

    def output_from_model(self, model, paths):
        return EmceeOutput(model=model, paths=paths)


class EmceeOutput(MCMCOutput):
    def __init__(
        self,
        model,
        paths,
        auto_correlation_check_size,
        auto_correlation_required_length,
        auto_correlation_change_threshold,
    ):

        super(EmceeOutput, self).__init__(model=model, paths=paths)

        self.auto_correlation_check_size = auto_correlation_check_size
        self.auto_correlation_required_length = auto_correlation_required_length
        self.auto_correlation_change_threshold = auto_correlation_change_threshold

    @property
    def backend(self):
        if os.path.isfile(self.paths.sym_path + "/emcee.hdf"):
            return emcee.backends.HDFBackend(
                filename=self.paths.sym_path + "/emcee.hdf"
            )
        else:
            raise FileNotFoundError(
                "The file emcee.hdf does not exist at the path " + self.paths.path
            )

    @property
    def pdf(self):
        import getdist

        return getdist.mcsamples.MCSamples(samples=self.samples_after_burn_in)

    @property
    def pdf_converged(self):
        return True

    @property
    def samples_after_burn_in(self):

        discard = int(3.0 * np.max(self.auto_correlation_times_of_parameters))
        thin = int(np.max(self.auto_correlation_times_of_parameters) / 2.0)
        return self.backend.get_chain(discard=discard, thin=thin, flat=True)

    @property
    def auto_correlation_times_of_parameters(self):
        return self.backend.get_autocorr_time(tol=0)

    @property
    def previous_auto_correlation_times_of_parameters(self):
        return emcee.autocorr.integrated_time(
            x=self.backend.get_chain()[: -self.auto_correlation_check_size, :, :], tol=0
        )

    @property
    def relative_auto_correlation_times(self):
        return (
            np.abs(
                self.previous_auto_correlation_times_of_parameters
                - self.auto_correlation_times_of_parameters
            )
            / self.auto_correlation_times_of_parameters
        )

    @property
    def converged(self):
        converged = np.all(
            self.auto_correlation_times_of_parameters
            * self.auto_correlation_required_length
            < self.total_samples
        )
        if converged:
            try:
                converged &= np.all(
                    self.relative_auto_correlation_times
                    < self.auto_correlation_change_threshold
                )
            except IndexError:
                return False
        return converged

    @property
    def total_samples(self):
        return len(self.backend.get_chain(flat=True))

    @property
    def total_walkers(self):
        return len(self.backend.get_chain()[0, :, 0])

    @property
    def total_steps(self):
        return len(self.backend.get_log_prob())

    @property
    def most_likely_index(self):
        return np.argmax(self.backend.get_log_prob(flat=True))

    @property
    def most_probable_model_parameters(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt' file which nlo from a \
        multinest lens.

        This file stores the parameters of the most probable model in the first half of entries and the most likely
        model in the second half of entries. The offset parameter is used to start at the desired model.

        """
        samples = self.samples_after_burn_in
        return [
            float(np.percentile(samples[:, i], [50]))
            for i in range(self.model.prior_count)
        ]

    @property
    def most_likely_model_parameters(self):
        """
        Read the most probable or most likely model values from the 'obj_summary.txt' file which nlo from a \
        multinest lens.

        This file stores the parameters of the most probable model in the first half of entries and the most likely
        model in the second half of entries. The offset parameter is used to start at the desired model.
        """
        return self.backend.get_chain(flat=True)[self.most_likely_index]

    @property
    def maximum_log_likelihood(self):
        return self.backend.get_log_prob(flat=True)[self.most_likely_index]

    def model_parameters_at_sigma_limit(self, sigma_limit):

        limit = math.erf(0.5 * sigma_limit * math.sqrt(2))

        samples = self.samples_after_burn_in

        return [
            tuple(np.percentile(samples[:, i], [100.0 * (1.0 - limit), 100.0 * limit]))
            for i in range(self.model.prior_count)
        ]

    def sample_model_parameters_from_sample_index(self, sample_index):
        """From a sample return the model parameters.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return list(self.pdf.samples[sample_index])

    def sample_weight_from_sample_index(self, sample_index):
        """From a sample return the sample weight.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return self.pdf.weights[sample_index]

    def sample_likelihood_from_sample_index(self, sample_index):
        """From a sample return the likelihood.

        NOTE: GetDist reads the log likelihood from the weighted_sample.txt file (column 2), which are defined as \
        -2.0*likelihood. This routine converts these back to likelihood.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return self.backend.get_log_prob(flat=True)[sample_index]

    def output_pdf_plots(self):

        pass



import pytest

import autofit as af
from test_autofit import mock


@pytest.fixture(name="phase")
def make_phase():
    phase = af.AbstractPhase(phase_name="phase name")
    phase.model.one = af.PriorModel(mock.Galaxy, light=mock.EllipticalLP)
    return phase


@pytest.fixture(name="collection")
def make_collection():
    collection = af.ResultsCollection()
    model = af.ModelMapper()
    model.one = af.PriorModel(mock.Galaxy, light=mock.EllipticalLP)
    instance = af.ModelInstance()
    instance.one = mock.Galaxy(light=mock.EllipticalLP())

    result = mock.Result(model=model, instance=instance)

    model = af.ModelMapper()
    instance = af.ModelInstance()

    model.hyper_galaxy = mock.HyperGalaxy
    instance.hyper_galaxy = mock.HyperGalaxy()

    hyper_result = mock.Result(model=model, instance=instance)

    result.hyper_result = hyper_result

    collection.add("phase name", result)

    return collection



import pytest

import autofit as af
from autofit import DeferredArgument


@pytest.fixture(name="uniform_dict")
def make_uniform_dict():
    return {"type": "Uniform", "lower_limit": 2, "upper_limit": 3}


@pytest.fixture(name="uniform_prior")
def make_uniform_prior(uniform_dict):
    return af.Prior.from_dict(uniform_dict)


@pytest.fixture(name="log_uniform_dict")
def make_log_uniform_dict():
    return {"type": "LogUniform", "lower_limit": 0.2, "upper_limit": 0.3}


@pytest.fixture(name="log_uniform_prior")
def make_log_uniform_prior(log_uniform_dict):
    return af.Prior.from_dict(log_uniform_dict)


@pytest.fixture(name="gaussian_dict")
def make_gaussian_dict():
    return {
        "type": "Gaussian",
        "lower_limit": -10,
        "upper_limit": 10,
        "mean": 3,
        "sigma": 4,
    }


@pytest.fixture(name="gaussian_prior")
def make_gaussian_prior(gaussian_dict):
    return af.Prior.from_dict(gaussian_dict)


@pytest.fixture(name="relative_width_dict")
def make_relative_width_dict():
    return {"type": "Relative", "value": 1.0}


@pytest.fixture(name="absolute_width_dict")
def make_absolute_width_dict():
    return {"type": "Absolute", "value": 2.0}


@pytest.fixture(name="relative_width_modifier")
def make_relative_width_modifier(relative_width_dict):
    return af.WidthModifier.from_dict(relative_width_dict)


@pytest.fixture(name="absolute_width_modifier")
def make_absolute_width_modifier(absolute_width_dict):
    return af.WidthModifier.from_dict(absolute_width_dict)


class TestWidth:
    def test_relative(self, relative_width_modifier):
        assert isinstance(relative_width_modifier, af.RelativeWidthModifier)
        assert relative_width_modifier.value == 1.0

    def test_absolute(self, absolute_width_modifier):
        assert isinstance(absolute_width_modifier, af.AbsoluteWidthModifier)
        assert absolute_width_modifier.value == 2.0


class TestDict:
    def test_uniform(self, uniform_prior, uniform_dict):
        assert uniform_prior.dict == uniform_dict

    def test_log_uniform(self, log_uniform_prior, log_uniform_dict):
        assert log_uniform_prior.dict == log_uniform_dict

    def test_gaussian(self, gaussian_prior, gaussian_dict):
        assert gaussian_prior.dict == gaussian_dict


class TestFromDict:
    def test_uniform(self, uniform_prior):
        assert isinstance(uniform_prior, af.UniformPrior)
        assert uniform_prior.lower_limit == 2
        assert uniform_prior.upper_limit == 3

    def test_log_uniform(self, log_uniform_prior, absolute_width_modifier):
        assert isinstance(log_uniform_prior, af.LogUniformPrior)
        assert log_uniform_prior.lower_limit == 0.2
        assert log_uniform_prior.upper_limit == 0.3

    def test_gaussian(self, gaussian_prior):
        assert isinstance(gaussian_prior, af.GaussianPrior)
        assert gaussian_prior.lower_limit == -10
        assert gaussian_prior.upper_limit == 10
        assert gaussian_prior.mean == 3
        assert gaussian_prior.sigma == 4

    def test_constant(self):
        result = af.Prior.from_dict({"type": "Constant", "value": 1.5})
        assert result == 1.5

    def test_deferred(self):
        result = af.Prior.from_dict({"type": "Deferred"})
        assert isinstance(result, DeferredArgument)
